"""
Module for reading/writing Neo objects in MATLAB format (.mat) versions
5 to 7.2.

This module is a bridge for MATLAB users who want to adopt the Neo object
representation. The nomenclature is the same but using Matlab structs and cell
arrays. With this module MATLAB users can use neo.io to read a format and
convert it to .mat.

Supported : Read/Write

Author: sgarcia, Robert Pröpper
"""

from collections.abc import Mapping
from datetime import datetime
import re

import numpy as np
import quantities as pq

from packaging.version import Version


from neo.io.baseio import BaseIO
from neo.core import (
    Block,
    Segment,
    AnalogSignal,
    IrregularlySampledSignal,
    Event,
    Epoch,
    SpikeTrain,
    Group,
    ImageSequence,
    ChannelView,
    RectangularRegionOfInterest,
    CircularRegionOfInterest,
    PolygonRegionOfInterest,
    objectnames,
    class_by_name,
    NeoReadWriteError,
)
from neo.core.regionofinterest import RegionOfInterest
from neo.core.baseneo import _container_name


def get_classname_from_container_name(container_name, struct):
    if container_name == "regionsofinterest":
        if "radius" in struct._fieldnames:
            return "CircularRegionOfInterest"
        elif "vertices" in struct._fieldnames:
            return "PolygonRegionOfInterest"
        else:
            return "RectangularRegionOfInterest"
    else:
        for classname in objectnames:
            if _container_name(classname) == container_name:
                return classname


PY_NONE = "Py_None"


class NeoMatlabIO(BaseIO):
    """
    Class for reading/writing Neo objects in MATLAB format (.mat) versions
    5 to 7.2.

    This module is a bridge for MATLAB users who want to adopt the Neo object
    representation.  The nomenclature is the same but using Matlab structs and
    cell arrays. With this module MATLAB users can use neo.io to read a format
    and convert it to .mat.

    Rules of conversion:
      * Neo classes are converted to MATLAB structs.
        e.g., a Block is a struct with attributes "name", "file_datetime", ...
      * Neo one_to_many relationships are cellarrays in MATLAB.
        e.g., ``seg.analogsignals[2]`` in Python Neo will be
        ``seg.analogsignals{3}`` in MATLAB.
      * Quantity attributes are represented by 2 fields in MATLAB.
        e.g., ``anasig.t_start = 1.5 * s`` in Python
        will be ``anasig.t_start = 1.5`` and ``anasig.t_start_unit = 's'``
        in MATLAB.
      * classes that inherit from Quantity (AnalogSignal, SpikeTrain, ...) in
        Python will have 2 fields (array and units) in the MATLAB struct.
        e.g.: ``AnalogSignal( [1., 2., 3.], 'V')`` in Python will be
        ``anasig.array = [1. 2. 3]`` and ``anasig.units = 'V'`` in MATLAB.

    1 - **Scenario 1: create data in MATLAB and read them in Python**

        This MATLAB code generates a block::

            block = struct();
            block.segments = { };
            block.name = 'my block with matlab';
            for s = 1:3
                seg = struct();
                seg.name = strcat('segment ',num2str(s));

                seg.analogsignals = { };
                for a = 1:5
                    anasig = struct();
                    anasig.signal = rand(100,1);
                    anasig.signal_units = 'mV';
                    anasig.t_start = 0;
                    anasig.t_start_units = 's';
                    anasig.sampling_rate = 100;
                    anasig.sampling_rate_units = 'Hz';
                    seg.analogsignals{a} = anasig;
                end

                seg.spiketrains = { };
                for t = 1:7
                    sptr = struct();
                    sptr.times = rand(30,1)*10;
                    sptr.times_units = 'ms';
                    sptr.t_start = 0;
                    sptr.t_start_units = 'ms';
                    sptr.t_stop = 10;
                    sptr.t_stop_units = 'ms';
                    seg.spiketrains{t} = sptr;
                end

                event = struct();
                event.times = [0, 10, 30];
                event.times_units = 'ms';
                event.labels = ['trig0'; 'trig1'; 'trig2'];
                seg.events{1} = event;

                epoch = struct();
                epoch.times = [10, 20];
                epoch.times_units = 'ms';
                epoch.durations = [4, 10];
                epoch.durations_units = 'ms';
                epoch.labels = ['a0'; 'a1'];
                seg.epochs{1} = epoch;

                block.segments{s} = seg;

            end

            save 'myblock.mat' block -V7


        This code reads it in Python::

            import neo
            r = neo.io.NeoMatlabIO(filename='myblock.mat')
            bl = r.read_block()
            print bl.segments[1].analogsignals[2]
            print bl.segments[1].spiketrains[4]


    2 - **Scenario 2: create data in Python and read them in MATLAB**

        This Python code generates the same block as in the previous scenario::

            import neo
            import quantities as pq
            from scipy import rand, array

            bl = neo.Block(name='my block with neo')
            for s in range(3):
                seg = neo.Segment(name='segment' + str(s))
                bl.segments.append(seg)
                for a in range(5):
                    anasig = neo.AnalogSignal(rand(100)*pq.mV, t_start=0*pq.s,
                                              sampling_rate=100*pq.Hz)
                    seg.analogsignals.append(anasig)
                for t in range(7):
                    sptr = neo.SpikeTrain(rand(40)*pq.ms, t_start=0*pq.ms, t_stop=10*pq.ms)
                    seg.spiketrains.append(sptr)
                ev = neo.Event([0, 10, 30]*pq.ms, labels=array(['trig0', 'trig1', 'trig2']))
                ep = neo.Epoch([10, 20]*pq.ms, durations=[4, 10]*pq.ms, labels=array(['a0', 'a1']))
                seg.events.append(ev)
                seg.epochs.append(ep)

            from neo.io.neomatlabio import NeoMatlabIO
            w = NeoMatlabIO(filename='myblock.mat')
            w.write_block(bl)


        This MATLAB code reads it::

            load 'myblock.mat'
            block.name
            block.segments{2}.analogsignals{3}.signal
            block.segments{2}.analogsignals{3}.signal_units
            block.segments{2}.analogsignals{3}.t_start
            block.segments{2}.analogsignals{3}.t_start_units


    3 - **Scenario 3: conversion**

        This Python code converts a Spike2 file to MATLAB::

            from neo import Block
            from neo.io import Spike2IO, NeoMatlabIO

            r = Spike2IO(filename='spike2.smr')
            w = NeoMatlabIO(filename='convertedfile.mat')
            blocks = r.read()
            w.write(blocks[0])

    """

    is_readable = True
    is_writable = True

    supported_objects = [
        Block,
        Segment,
        AnalogSignal,
        IrregularlySampledSignal,
        Epoch,
        Event,
        SpikeTrain,
        Group,
        ImageSequence,
        ChannelView,
        RectangularRegionOfInterest,
        CircularRegionOfInterest,
        PolygonRegionOfInterest,
    ]
    readable_objects = [Block]
    writeable_objects = [Block]

    has_header = False
    is_streameable = False
    read_params = {Block: []}
    write_params = {Block: []}

    name = "neomatlab"
    extensions = ["mat"]

    mode = "file"

    def __init__(self, filename=None):
        """
        This class read/write neo objects in matlab 5 to 7.2 format.

        Arguments:
            filename : the filename to read
        """
        import scipy.version

        if Version(scipy.version.version) < Version("0.12.0"):
            raise ImportError(
                "your scipy version is too old to support "
                + "MatlabIO, you need at least 0.12.0. "
                + f"You have {scipy.version.version}"
            )

        BaseIO.__init__(self)
        self.filename = filename
        self._refs = {}

    def read_block(self, lazy=False):
        """
        Arguments:

        """
        import scipy.io

        if lazy:
            raise NeoReadWriteError(f"This IO does not support lazy reading")

        d = scipy.io.loadmat(self.filename, struct_as_record=False, squeeze_me=True, mat_dtype=True)
        if "block" not in d:
            self.logger.exception("No block in " + self.filename)
            return None

        bl_struct = d["block"]
        bl = self.create_ob_from_struct(bl_struct, "Block")
        self._resolve_references(bl)
        bl.check_relationships()
        return bl

    def write_block(self, bl, **kargs):
        """
        Arguments:
            bl: the block to be saved
        """
        import scipy.io

        bl_struct = self.create_struct_from_obj(bl)

        for seg in bl.segments:
            seg_struct = self.create_struct_from_obj(seg)
            bl_struct["segments"].append(seg_struct)

            for container_name in seg._child_containers:
                for child_obj in getattr(seg, container_name):
                    child_struct = self.create_struct_from_obj(child_obj)
                    seg_struct[container_name].append(child_struct)

        for group in bl.groups:
            group_structure = self.create_struct_from_obj(group)
            bl_struct["groups"].append(group_structure)

            for container_name in group._child_containers:
                for child_obj in getattr(group, container_name):
                    if isinstance(child_obj, (ChannelView, RegionOfInterest)):
                        child_struct = self.create_struct_from_view(child_obj)
                        group_structure[container_name].append(child_struct)
                    else:
                        group_structure[container_name].append(id(child_obj))

        scipy.io.savemat(self.filename, {"block": bl_struct}, oned_as="row")

    def _get_matlab_value(self, ob, attrname):
        units = None
        if hasattr(ob, "_quantity_attr") and ob._quantity_attr == attrname:
            units = ob.dimensionality.string
            value = ob.magnitude
        else:
            try:
                value = getattr(ob, attrname)
            except AttributeError:
                value = ob[attrname]
            if isinstance(value, pq.Quantity):
                units = value.dimensionality.string
                value = value.magnitude
            elif isinstance(value, datetime):
                value = str(value)
            elif isinstance(value, Mapping):
                new_value = {}
                for key in value:
                    subvalue, subunits = self._get_matlab_value(value, key)
                    if subvalue is not None:
                        new_value[key] = subvalue
                        if subunits:
                            new_value[f"{key}_units"] = subunits
                    elif attrname == "annotations":
                        # In general we don't send None to MATLAB
                        # but we make an exception for annotations.
                        # However, we have to save then retrieve some
                        # special value as actual `None` is ignored by default.
                        new_value[key] = PY_NONE
                value = new_value
        return value, units

    def create_struct_from_obj(self, ob):
        struct = {"neo_id": id(ob)}

        # relationship
        for childname in getattr(ob, "_child_containers", []):
            supported_containers = [_container_name(subob.__name__) for subob in self.supported_objects]
            if childname in supported_containers:
                struct[childname] = []

        # attributes
        all_attrs = list(ob._all_attrs)
        if hasattr(ob, "annotations"):
            all_attrs.append(("annotations", type(ob.annotations)))

        for attr in all_attrs:
            attrname, attrtype = attr[0], attr[1]
            attr_value, attr_units = self._get_matlab_value(ob, attrname)
            if attr_value is not None:
                struct[attrname] = attr_value
                if attr_units:
                    struct[attrname + "_units"] = attr_units
        return struct

    def create_struct_from_view(self, ob):
        # for "view" objects (ChannelView and RegionOfInterest), we just store
        # a reference to the object (AnalogSignal, ImageSequence) that the view
        # points to
        struct = self.create_struct_from_obj(ob)
        obj_name = ob._necessary_attrs[0][0]  # this is fragile, better to add an attribute _view_attr
        viewed_obj = getattr(ob, obj_name)
        struct[obj_name] = id(viewed_obj)
        struct["viewed_classname"] = viewed_obj.__class__.__name__
        return struct

    def create_ob_from_struct(self, struct, classname):
        cl = class_by_name[classname]

        # ~ if is_quantity:
        if hasattr(cl, "_quantity_attr"):
            quantity_attr = cl._quantity_attr
            arr = getattr(struct, quantity_attr)
            # ~ data_complement = dict(units=str(struct.units))
            data_complement = dict(units=str(getattr(struct, quantity_attr + "_units")))
            if "sampling_rate" in (at[0] for at in cl._necessary_attrs):
                # put fake value for now, put correct value later
                data_complement["sampling_rate"] = 0 * pq.kHz
            try:
                len(arr)
            except TypeError:
                # strange scipy.io behavior: if len is 1 we get a float
                arr = np.array(arr)
                arr = arr.reshape((-1,))  # new view with one dimension
            if "t_stop" in (at[0] for at in cl._necessary_attrs):
                if len(arr) > 0:
                    data_complement["t_stop"] = arr.max()
                else:
                    data_complement["t_stop"] = 0.0
            if "t_start" in (at[0] for at in cl._necessary_attrs):
                if len(arr) > 0:
                    data_complement["t_start"] = arr.min()
                else:
                    data_complement["t_start"] = 0.0
            if "spatial_scale" in (at[0] for at in cl._necessary_attrs):
                if len(arr) > 0:
                    data_complement["spatial_scale"] = arr
                else:
                    data_complement["spatial_scale"] = 1.0

            if "times" in (at[0] for at in cl._necessary_attrs) and quantity_attr != "times":
                # handle IrregularlySampledSignal
                times = getattr(struct, "times")
                data_complement["time_units"] = getattr(struct, "times_units")
                ob = cl(times, arr, **data_complement)
            else:
                ob = cl(arr, **data_complement)
        elif cl.is_view:
            kwargs = {}
            for i, attr in enumerate(cl._necessary_attrs):
                value = getattr(struct, attr[0])
                if i == 0:
                    # this is a bit hacky, should really add an attribute _view_attr to ChannelView and RegionOfInterest
                    if not isinstance(value, int):  # object id
                        raise TypeError(f"value must be int not of type {type(value)}")
                    kwargs[attr[0]] = _Ref(identifier=value, target_class_name=struct.viewed_classname)
                else:
                    if attr[1] == np.ndarray and isinstance(value, int):
                        value = np.array([value])
                    kwargs[attr[0]] = value
            ob = cl(**kwargs)
        else:
            ob = cl()

        for attrname in struct._fieldnames:
            # check children
            if attrname in getattr(ob, "_child_containers", []):
                child_struct = getattr(struct, attrname)
                try:
                    # try must only surround len() or other errors are captured
                    child_len = len(child_struct)
                except TypeError:
                    # strange scipy.io behavior: if len is 1 there is no len()
                    child_struct = [child_struct]
                    child_len = 1

                for c in range(child_len):
                    child_class_name = get_classname_from_container_name(attrname, child_struct[c])
                    if classname == "Group":
                        if child_class_name == ("ChannelView") or "RegionOfInterest" in child_class_name:
                            child = self.create_ob_from_struct(child_struct[c], child_class_name)
                        else:
                            child = _Ref(child_struct[c], child_class_name)
                    else:
                        child = self.create_ob_from_struct(child_struct[c], child_class_name)
                    getattr(ob, attrname.lower()).append(child)
                continue

            # attributes
            if attrname.endswith("_units") or attrname == "units":
                # linked with another field
                continue

            if hasattr(cl, "_quantity_attr") and cl._quantity_attr == attrname:
                continue

            if cl.is_view and attrname in (
                "obj",
                "index",
                "image_sequence",
                "x",
                "y",
                "radius",
                "width",
                "height",
                "vertices",
            ):
                continue

            item = getattr(struct, attrname)
            attributes = cl._necessary_attrs + cl._recommended_attrs + (("annotations", dict),)
            dict_attributes = dict([(a[0], a[1:]) for a in attributes])

            if attrname in dict_attributes:
                attrtype = dict_attributes[attrname][0]
                if attrtype == datetime:
                    m = r"(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+).(\d+)"
                    r = re.findall(m, str(item))
                    if len(r) == 1:
                        item = datetime(*[int(e) for e in r[0]])
                    else:
                        item = None
                elif attrtype == np.ndarray:
                    dt = dict_attributes[attrname][2]
                    try:
                        item = item.astype(dt)
                    except AttributeError:
                        # it seems arrays of length 1 are stored as scalars
                        item = np.array([item], dtype=dt)
                elif attrtype == pq.Quantity:
                    ndim = dict_attributes[attrname][1]
                    units = str(getattr(struct, attrname + "_units"))
                    if ndim == 0:
                        item = pq.Quantity(item, units)
                    else:
                        item = pq.Quantity(item, units)
                elif attrtype == dict:
                    new_item = {}
                    for fn in item._fieldnames:
                        value = getattr(item, fn)
                        if value == PY_NONE:
                            value = None
                        new_item[fn] = value
                    item = new_item
                else:
                    item = attrtype(item)

            setattr(ob, attrname, item)

        neo_id = getattr(struct, "neo_id", None)
        if neo_id:
            setattr(ob, "_id", neo_id)
        return ob

    def _resolve_references(self, bl):
        if bl.groups:
            obj_lookup = {}
            for ob in bl.children_recur:
                if hasattr(ob, "_id"):
                    obj_lookup[ob._id] = ob
            for grp in bl.groups:
                for container_name in grp._child_containers:
                    container = getattr(grp, container_name)
                    for i, item in enumerate(container):
                        if isinstance(item, _Ref):
                            if not isinstance(item.identifier, (int, np.integer)):
                                raise TypeError(
                                    f"item.identifier must be either int or np.integer not of type {type(item.identifier)}"
                                )
                            # A reference to an object that already exists
                            container[i] = obj_lookup[item.identifier]
                        else:
                            # ChannelView and RegionOfInterest
                            if not item.is_view:
                                raise TypeError(f"`item` must be a view")
                            if not isinstance(item.obj, _Ref):
                                raise TypeError(f"`item.obj` must be a {_Ref} and is of type {type(item.obj)}")
                            item.obj = obj_lookup[item.obj.identifier]


class _Ref:
    def __init__(self, identifier, target_class_name):
        self.identifier = identifier
        if target_class_name:
            self.target_cls = class_by_name[target_class_name]
        else:
            self.target_cls = None

    @property
    def proxy_for(self):
        return self.target_cls

    @property
    def data_children_recur(self):
        return []

    @property
    def container_children_recur(self):
        return []



"""
AxographRawIO
=============

RawIO class for reading AxoGraph files (.axgd, .axgx)

Original author: Jeffrey Gill

Documentation of the AxoGraph file format provided by the developer is
incomplete and in some cases incorrect. The primary sources of official
documentation are found in two out-of-date documents:

    - AxoGraph X User Manual, provided with AxoGraph and also available online:
        https://axograph.com/documentation/AxoGraph%20User%20Manual.pdf

    - AxoGraph_ReadWrite.h, a header file that is part of a C++ program
      provided with AxoGraph, and which is also available here:
        https://github.com/CWRUChielLab/axographio/blob/master/
            axographio/include/axograph_readwrite/AxoGraph_ReadWrite.h

These were helpful starting points for building this RawIO, especially for
reading the beginnings of AxoGraph files, but much of the rest of the file
format was deciphered by reverse engineering and guess work. Some portions of
the file format remain undeciphered.

The AxoGraph file format is versatile in that it can represent both time series
data collected during data acquisition and non-time series data generated
through manual or automated analysis, such as power spectrum analysis. This
implementation of an AxoGraph file format reader makes no effort to decoding
non-time series data. For simplicity, it makes several assumptions that should
be valid for any file generated directly from episodic or continuous data
acquisition without significant post-acquisition modification by the user.

Detailed logging is provided during header parsing for debugging purposes:
    >>> import logging
    >>> r = AxographRawIO(filename)
    >>> r.logger.setLevel(logging.DEBUG)
    >>> r.parse_header()

Background and Terminology
--------------------------

Acquisition modes:
    AxoGraph can operate in two main data acquisition modes:
    - Episodic "protocol-driven" acquisition mode, in which the program records
      from specified signal channels for a fixed duration each time a trigger
      is detected. Each trigger-activated recording is called an "episode".
      From files acquired in this mode, AxographRawIO creates multiple Neo
      Segments, one for each episode, unless force_single_segment=True.
    - Continuous "chart recorder" acquisition mode, in which it creates a
      continuous recording that can be paused and continued by the user
      whenever they like. From files acquired in this mode, AxographRawIO
      creates a single Neo Segment.

"Episode": analogous to a Neo Segment
    See descriptions of acquisition modes above and of groups below.

"Column": analogous to a Quantity array
    A column is a 1-dimensional array of data, stored in any one of a number of
    data types (e.g., scaled ints or floats). In the oldest version of the
    AxoGraph file format, even time was stored as a 1-dimensional array. In
    newer versions, time is stored as a special type of "column" that is really
    just a starting time and a sampling period.

    Column data appears in series in the file, i.e., all of the first column's
    data appears before the second column's. As an aside, because of this
    design choice AxoGraph cannot write data to disk as it is collected but
    must store it all in memory until data acquisition ends. This also affected
    how file slicing was implemented for this RawIO: Instead of using a single
    memmap to address into a 2-dimensional block of data, AxographRawIO
    constructs multiple 1-dimensional memmaps, one for each column, each with
    its own offset.

    Each column's data array is preceded by a header containing the column
    title, which normally contains the units (e.g., "Current (nA)"). Data
    recorded in episodic acquisition mode will contain a repeating sequence of
    column names, where each repetition corresponds to an episode (e.g.,
    "Time", "Column A", "Column B", "Column A", "Column B", etc.).

    AxoGraph offers a spreadsheet view for viewing all column data.

"Trace": analogous to a single-channel Neo AnalogSignal
    A trace is a 2-dimensional series. Raw data is not stored in the part of
    the file concerned with traces. Instead, in the header for each trace are
    indexes pointing to two data columns, defined earlier in the file,
    corresponding to the trace's x and y data. These indexes can be changed in
    AxoGraph under the "Assign X and Y Columns" tool, though doing so may
    violate assumptions made by AxographRawIO.

    For time series data collected under the usual data acquisition modes that
    has not been modified after collection by the user, the x-index always
    points to the time column; one trace exists for each non-time column, with
    the y-index pointing to that column.

    Traces are analogous to AnalogSignals in Neo. However, for simplicity of
    implementation, AxographRawIO does not actually check the pairing of
    columns in the trace headers. Instead it assumes the default pairing
    described above when it creates signal channels while scanning through
    columns. Older versions of the AxoGraph file format lack trace headers
    entirely, so this is the most general solution.

    Trace headers contain additional information about the series, such as plot
    style, which is parsed by AxographRawIO and made available in
    self.info['trace_header_info_list'] but is otherwise unused.

"Group": analogous to a Neo ChannelIndex for matching channels across Segments
    A group is a collection of one or more traces. Like traces, raw data is not
    stored in the part of the file concerned with groups. Instead, each trace
    header contains an index pointing to the group it is assigned to. Group
    assignment of traces can be changed in AxoGraph under the "Group Traces"
    tool, or by using the "Merge Traces" or "Separate Traces" commands, though
    doing so may violate assumptions made by AxographRawIO.

    Files created in episodic acquisition mode contain multiple traces per
    group, one for each episode. In that mode, a group corresponds to a signal
    channel and is analogous to a ChannelIndex in Neo; the traces within the
    group represent the time series recorded for that channel across episodes
    and are analogous to AnalogSignals from multiple Segments in Neo.

    In contrast, files created in continuous acquisition mode contain one trace
    per group, each corresponding to a signal channel. In that mode, groups and
    traces are basically conceptually synonymous, though the former can still
    be thought of as analogous to ChannelIndexes in Neo for a single-Segment.

    Group headers are only consulted by AxographRawIO to determine if is safe
    to interpret a file as episodic and therefore translatable to multiple
    Segments in Neo. Certain criteria have to be met, such as all groups
    containing equal numbers of traces and each group having homogeneous signal
    parameters. If trace grouping was modified by the user after data
    acquisition, this may result in the file being interpreted as
    non-episodic. Older versions of the AxoGraph file format lack group headers
    entirely, so these files are never deemed safe to interpret as episodic,
    even if the column names follow a repeating sequence as described above.

"Tag" / "Event marker": analogous to a Neo Event
    In continuous acquisition mode, the user can press a hot key to tag a
    moment in time with a short label. Additionally, if the user stops or
    restarts data acquisition in this mode, a tag is created automatically with
    the label "Stop" or "Start", respectively. These are displayed by AxoGraph
    as event markers. AxographRawIO will organize all event markers into a
    single Neo Event channel with the name "AxoGraph Tags".

    In episodic acquisition mode, the tag hot key behaves differently. The
    current episode number is recorded in a user-editable notes section of the
    file, made available by AxographRawIO in self.info['notes']. Because these
    do not correspond to moments in time, they are not processed into Neo
    Events.

"Interval bar": analogous to a Neo Epoch
    After data acquisition, the user can annotate an AxoGraph file with
    horizontal, labeled bars called interval bars that span a specified period
    of time. These are not episode specific. AxographRawIO will organize all
    interval bars into a single Neo Epoch channel with the name "AxoGraph
    Intervals".
"""

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

import os
from datetime import datetime
from io import open, BufferedReader
from struct import unpack, calcsize

import numpy as np


class AxographRawIO(BaseRawIO):
    """
    RawIO class for reading AxoGraph files (.axgd, .axgx)

    Parameters
    ----------
    filename: str
        File name of the AxoGraph file to read.
    force_single_segment: bool, default: False
        Episodic files are normally read as multi-Segment Neo objects. This
        parameter can force AxographRawIO to put all signals into a single
        Segment.

    Examples
    --------
        >>> import neo
        >>> reader = neo.rawio.AxographRawIO(filename=filename)
        >>> reader.parse_header()
        >>> print(reader)

        >>> # get signals
        >>> raw_chunk = reader.get_analogsignal_chunk(block_index=0,
        ...                                           seg_index=0,
        ...                                           i_start=0,
        ...                                           i_stop=1024,
        ...                                           channel_names=channel_names)

        >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk,
        ...                                             dtype='float64',
        ...                                             channel_names=channel_names)
        >>> print(float_chunk)

        >>> # get event markers
        >>> ev_raw_times, _, ev_labels = reader.get_event_timestamps(event_channel_index=0)
        >>> ev_times = reader.rescale_event_timestamp(ev_raw_times, dtype='float64')
        >>> print([ev for ev in zip(ev_times, ev_labels)])

        >>> # get interval bars
        >>> ep_raw_times, ep_raw_durations, ep_labels = reader.get_event_timestamps(event_channel_index=1)
        >>> ep_times = reader.rescale_event_timestamp(ep_raw_times, dtype='float64')
        >>> ep_durations = reader.rescale_epoch_duration(ep_raw_durations, dtype='float64')
        >>> print([ep for ep in zip(ep_times, ep_durations, ep_labels)])

        >>> # get notes
        >>> print(reader.info['notes'])

        >>> # get other miscellaneous info
        >>> print(reader.info)
    """

    name = "AxographRawIO"
    description = "This IO reads .axgd/.axgx files created with AxoGraph"
    extensions = ["axgd", "axgx", ""]
    rawmode = "one-file"

    def __init__(self, filename, force_single_segment=False):
        BaseRawIO.__init__(self)
        self.filename = filename
        self.force_single_segment = force_single_segment

    def _parse_header(self):

        self.header = {}

        self._scan_axograph_file()

        if not self.force_single_segment and self._safe_to_treat_as_episodic():
            self.logger.debug("Will treat as episodic")
            self._convert_to_multi_segment()
        else:
            self.logger.debug("Will not treat as episodic")
        self.logger.debug("")

        self._generate_minimal_annotations()
        blk_annotations = self.raw_annotations["blocks"][0]
        blk_annotations["format_ver"] = self.info["format_ver"]
        blk_annotations["comment"] = self.info["comment"] if "comment" in self.info else None
        blk_annotations["notes"] = self.info["notes"] if "notes" in self.info else None
        blk_annotations["rec_datetime"] = self._get_rec_datetime()

        # modified time is not ideal but less prone to
        # cross-platform issues than created time (ctime)
        blk_annotations["file_datetime"] = datetime.fromtimestamp(os.path.getmtime(self.filename))

    def _source_name(self):
        return self.filename

    def _segment_t_start(self, block_index, seg_index):
        # same for all segments
        return self._t_start

    def _segment_t_stop(self, block_index, seg_index):
        # same for all signals in all segments
        t_stop = self._t_start + len(self._raw_signals[seg_index][0]) * self._sampling_period
        return t_stop

    ###
    # signal and channel zone

    def _get_signal_size(self, block_index, seg_index, stream_index):
        # same for all signals in all segments
        return len(self._raw_signals[seg_index][0])

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        # same for all signals in all segments
        return self._t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        if channel_indexes is None or np.all(channel_indexes == slice(None, None, None)):
            channel_indexes = range(self.signal_channels_count(stream_index))

        raw_signals = [
            self._raw_signals[seg_index][channel_index][slice(i_start, i_stop)] for channel_index in channel_indexes
        ]
        raw_signals = np.array(raw_signals).T  # loads data into memory

        return raw_signals

    ###
    # spiketrain and unit zone

    def _spike_count(self, block_index, seg_index, unit_index):
        # not supported
        return None

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        # not supported
        return None

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        # not supported
        return None

    ###
    # spike waveforms zone

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        # not supported
        return None

    ###
    # event and epoch zone

    def _event_count(self, block_index, seg_index, event_channel_index):

        # Retrieve size of either event or epoch channel:
        #   event_channel_index: 0 AxoGraph Tags, 1 AxoGraph Intervals
        # AxoGraph tags can only be inserted in continuous data acquisition
        # mode. When the tag hot key is pressed in episodic acquisition mode,
        # the notes are updated with the current episode number instead of an
        # instantaneous event marker being created. This means that Neo-like
        # Events cannot be generated by AxoGraph for multi-Segment (episodic)
        # files. Furthermore, Neo-like Epochs (interval markers) are not
        # episode specific. For these reasons, this function ignores seg_index.

        return self._raw_event_epoch_timestamps[event_channel_index].size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):

        # Retrieve either event or epoch data, unscaled:
        #   event_channel_index: 0 AxoGraph Tags, 1 AxoGraph Intervals
        # AxoGraph tags can only be inserted in continuous data acquisition
        # mode. When the tag hot key is pressed in episodic acquisition mode,
        # the notes are updated with the current episode number instead of an
        # instantaneous event marker being created. This means that Neo-like
        # Events cannot be generated by AxoGraph for multi-Segment (episodic)
        # files. Furthermore, Neo-like Epochs (interval markers) are not
        # episode specific. For these reasons, this function ignores seg_index.

        timestamps = self._raw_event_epoch_timestamps[event_channel_index]
        durations = self._raw_event_epoch_durations[event_channel_index]
        labels = self._event_epoch_labels[event_channel_index]

        if durations is None:
            # events
            if t_start is not None:
                # keep if event occurs after t_start ...
                keep = timestamps >= int(t_start / self._sampling_period)
                timestamps = timestamps[keep]
                labels = labels[keep]

            if t_stop is not None:
                # ... and before t_stop
                keep = timestamps <= int(t_stop / self._sampling_period)
                timestamps = timestamps[keep]
                labels = labels[keep]
        else:
            # epochs
            if t_start is not None:
                # keep if epoch ends after t_start ...
                keep = timestamps + durations >= int(t_start / self._sampling_period)
                timestamps = timestamps[keep]
                durations = durations[keep]
                labels = labels[keep]

            if t_stop is not None:
                # ... and starts before t_stop
                keep = timestamps <= int(t_stop / self._sampling_period)
                timestamps = timestamps[keep]
                durations = durations[keep]
                labels = labels[keep]

        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        # Scale either event or epoch start times from sample index to seconds
        # (t_start shouldn't be added)
        event_times = event_timestamps.astype(dtype) * self._sampling_period
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        # Scale epoch durations from samples to seconds
        epoch_durations = raw_duration.astype(dtype) * self._sampling_period
        return epoch_durations

    ###
    # multi-segment zone

    def _safe_to_treat_as_episodic(self):
        """
        The purpose of this function is to determine if the file contains any
        irregularities in its grouping of traces such that it cannot be treated
        as episodic. Even "continuous" recordings can be treated as
        single-episode recordings and could be identified as safe by this
        function. Recordings in which the user has changed groupings to create
        irregularities should be caught by this function.
        """

        # First check: Old AxoGraph file formats do not contain enough metadata
        # to know for certain that the file is episodic.
        if self.info["format_ver"] < 3:
            self.logger.debug("Cannot treat as episodic because old format " "contains insufficient metadata")
            return False

        # Second check: If the file is episodic, it should report that it
        # contains more than 1 episode.
        if "n_episodes" not in self.info:
            self.logger.debug("Cannot treat as episodic because episode " "metadata is missing or could not be parsed")
            return False
        if self.info["n_episodes"] == 1:
            self.logger.debug("Cannot treat as episodic because file reports " "one episode")
            return False

        # Third check: If the file is episodic, groups of traces should all
        # contain the same number of traces, one for each episode. This is
        # generally true of "continuous" (single-episode) recordings as well,
        # which normally have 1 trace per group.
        if "group_header_info_list" not in self.info:
            self.logger.debug("Cannot treat as episodic because group " "metadata is missing or could not be parsed")
            return False
        if "trace_header_info_list" not in self.info:
            self.logger.debug("Cannot treat as episodic because trace " "metadata is missing or could not be parsed")
            return False

        group_id_to_col_indexes = {}
        for group_id in self.info["group_header_info_list"]:
            col_indexes = []
            for trace_header in self.info["trace_header_info_list"].values():
                if trace_header["group_id_for_this_trace"] == group_id:
                    col_indexes.append(trace_header["y_index"])
            group_id_to_col_indexes[group_id] = col_indexes
        n_traces_by_group = {k: len(v) for k, v in group_id_to_col_indexes.items()}
        all_groups_have_same_number_of_traces = len(np.unique(list(n_traces_by_group.values()))) == 1

        if not all_groups_have_same_number_of_traces:
            self.logger.debug("Cannot treat as episodic because groups differ " "in number of traces")
            return False

        # Fourth check: The number of traces in each group should equal
        # n_episodes.
        n_traces_per_group = np.unique(list(n_traces_by_group.values()))
        if n_traces_per_group != self.info["n_episodes"]:
            self.logger.debug(
                "Cannot treat as episodic because n_episodes " "does not match number of traces per group"
            )
            return False

        # Fifth check: If the file is episodic, all traces within a group
        # should have identical signal channel parameters (e.g., name, units)
        # except for their unique ids. This too is generally true of
        # "continuous" (single-episode) files, which normally have 1 trace per
        # group.
        signal_channels_with_ids_dropped = self.header["signal_channels"][
            [n for n in self.header["signal_channels"].dtype.names if n != "id"]
        ]
        group_has_uniform_signal_parameters = {}
        for group_id, col_indexes in group_id_to_col_indexes.items():
            # subtract 1 from indexes in next statement because time is not
            # included in signal_channels
            signal_params_for_group = np.array(signal_channels_with_ids_dropped[np.array(col_indexes) - 1])
            group_has_uniform_signal_parameters[group_id] = len(np.unique(signal_params_for_group)) == 1
        all_groups_have_uniform_signal_parameters = np.all(list(group_has_uniform_signal_parameters.values()))

        if not all_groups_have_uniform_signal_parameters:
            self.logger.debug("Cannot treat as episodic because some groups " "have heterogeneous signal parameters")
            return False

        # all checks passed
        self.logger.debug("Can treat as episodic")
        return True

    def _convert_to_multi_segment(self):
        """
        Reshape signal headers and signal data for an episodic file
        """

        self.header["nb_segment"] = [self.info["n_episodes"]]

        # drop repeated signal headers
        self.header["signal_channels"] = self.header["signal_channels"].reshape(self.info["n_episodes"], -1)[0]

        # reshape signal memmap list
        new_sig_memmaps = []
        n_channels = len(self.header["signal_channels"])
        sig_memmaps = self._raw_signals[0]
        for first_index in np.arange(0, len(sig_memmaps), n_channels):
            new_sig_memmaps.append(sig_memmaps[first_index : first_index + n_channels])
        self._raw_signals = new_sig_memmaps

        self.logger.debug(f"New number of segments: {self.info['n_episodes']}")

        return

    def _get_rec_datetime(self):
        """
        Determine the date and time at which the recording was started from
        automatically generated notes. How these notes should be parsed differs
        depending on whether the recording was obtained in episodic or
        continuous acquisition mode.
        """

        rec_datetime = None
        date_string = ""
        time_string = ""
        datetime_string = ""

        if "notes" not in self.info:
            return None

        for note_line in self.info["notes"].split("\n"):

            # episodic acquisition mode
            if note_line.startswith("Created on "):
                date_string = note_line.strip("Created on ")
            if note_line.startswith("Start data acquisition at "):
                time_string = note_line.strip("Start data acquisition at ")

            # continuous acquisition mode
            if note_line.startswith("Created : "):
                datetime_string = note_line.strip("Created : ")

        if date_string and time_string:
            datetime_string = " ".join([date_string, time_string])

        if datetime_string:
            try:
                rec_datetime = datetime.strptime(datetime_string, "%a %b %d %Y %H:%M:%S")
            except ValueError:
                pass

        return rec_datetime

    def _scan_axograph_file(self):
        """
        This function traverses the entire AxoGraph file, constructing memmaps
        for signals and collecting channel information and other metadata
        """

        self.info = {}

        with open(self.filename, "rb") as fid:
            f = StructFile(fid)

            self.logger.debug(f"filename: {self.filename}")
            self.logger.debug("")

            # the first 4 bytes are always a 4-character file type identifier
            # - for early versions of AxoGraph, this identifier was 'AxGr'
            # - starting with AxoGraph X, the identifier is 'axgx'
            header_id = f.read(4).decode("utf-8")
            self.info["header_id"] = header_id
            if header_id not in ["AxGr", "axgx"]:
                raise NeoReadWriteError(f'File "{self.filename}" is not an AxoGraph binary file! Use other neo reader')

            self.logger.debug(f"{header_id=}")

            # the next two numbers store the format version number and the
            # number of data columns to follow
            # - for 'AxGr' files, these numbers are 2-byte unsigned short ints
            # - for 'axgx' files, these numbers are 4-byte long ints
            # - the 4-character identifier changed from 'AxGr' to 'axgx' with
            #   format version 3
            if header_id == "AxGr":
                format_ver, n_cols = f.read_f("HH")
                if format_ver != 1 and format_ver != 2:
                    raise ValueError(
                        f'Mismatch between header identifier "{header_id}" and format ' f'version "{format_ver}"!'
                    )
            elif header_id == "axgx":
                format_ver, n_cols = f.read_f("ll")
                if format_ver < 3:
                    raise ValueError(
                        f'Mismatch between header identifier "{header_id}" and format ' f'version "{format_ver}"!'
                    )
            else:
                raise NotImplementedError(f'unimplemented file header identifier "{header_id}"!')
            self.info["format_ver"] = format_ver
            self.info["n_cols"] = n_cols

            self.logger.debug(f"format_ver: {format_ver}")
            self.logger.debug(f"n_cols: {n_cols}")
            self.logger.debug("")

            ##############################################
            # BEGIN COLUMNS

            sig_memmaps = []
            sig_channels = []
            for i in range(n_cols):

                self.logger.debug(f"== COLUMN INDEX {i} ==")

                ##############################################
                # NUMBER OF DATA POINTS IN COLUMN

                n_points = f.read_f("l")

                self.logger.debug(f"n_points: {n_points}")

                ##############################################
                # COLUMN TYPE

                # depending on the format version, data columns may have a type
                # - prior to version 3, column types did not exist and data was
                #   stored in a fixed pattern
                # - beginning with version 3, several data types are available
                #   as documented in AxoGraph_ReadWrite.h
                if format_ver == 1 or format_ver == 2:
                    col_type = None
                elif format_ver >= 3:
                    col_type = f.read_f("l")
                else:
                    raise NotImplementedError(f'unimplemented file format version "{format_ver}"!')

                self.logger.debug(f"col_type: {col_type}")

                ##############################################
                # COLUMN NAME AND UNITS

                # depending on the format version, column titles are stored
                # differently
                # - prior to version 3, column titles were stored as
                #   fixed-length 80-byte Pascal strings
                # - beginning with version 3, column titles are stored as
                #   variable-length strings (see StructFile.read_string for
                #   details)
                if format_ver == 1 or format_ver == 2:
                    title = f.read_f("80p").decode("utf-8")
                elif format_ver >= 3:
                    title = f.read_f("S")
                else:
                    raise NotImplementedError(f'unimplemented file format version "{format_ver}"!')

                self.logger.debug(f"title: {title}")

                # units are given in parentheses at the end of a column title,
                # unless units are absent
                if len(title.split()) > 0 and title.split()[-1][0] == "(" and title.split()[-1][-1] == ")":
                    name = " ".join(title.split()[:-1])
                    units = title.split()[-1].strip("()")
                else:
                    name = title
                    units = ""

                self.logger.debug(f"name: {name}")
                self.logger.debug(f"units: {units}")

                ##############################################
                # COLUMN DTYPE, SCALE, OFFSET

                if format_ver == 1:

                    # for format version 1, all columns are arrays of floats

                    dtype = "f"
                    gain, offset = 1, 0  # data is neither scaled nor off-set

                elif format_ver == 2:

                    # for format version 2, the first column is a "series" of
                    # regularly spaced values specified merely by a first value
                    # and an increment, and all subsequent columns are arrays
                    # of shorts with a scaling factor

                    if i == 0:

                        # series
                        first_value, increment = f.read_f("ff")

                        self.logger.debug(f"interval: {increment}, freq: {1 / increment}")
                        self.logger.debug(f"start: {first_value}, end: {first_value + increment * (n_points - 1)}")

                        # assume this is the time column
                        t_start, sampling_period = first_value, increment
                        self.info["t_start"] = t_start
                        self.info["sampling_period"] = sampling_period

                        self.logger.debug("")

                        continue  # skip memmap, chan info for time col

                    else:

                        # scaled short
                        dtype = "h"
                        gain, offset = f.read_f("f"), 0  # data is scaled without offset

                elif format_ver >= 3:

                    # for format versions 3 and later, the column type
                    # determines how the data should be read
                    # - column types 1, 2, 3, and 8 are not defined in
                    #   AxoGraph_ReadWrite.h
                    # - column type 9 is different from the others in that it
                    #   represents regularly spaced values
                    #   (such as times at a fixed frequency) specified by a
                    #   first value and an increment, without storing a large
                    #   data array

                    if col_type == 9:

                        # series
                        first_value, increment = f.read_f("dd")

                        self.logger.debug(f"interval: {increment}, freq: {1 / increment}")
                        self.logger.debug("start: {first_value}, end: {first_value + increment * (n_points - 1)}")

                        if i == 0:

                            # assume this is the time column
                            t_start, sampling_period = first_value, increment
                            self.info["t_start"] = t_start
                            self.info["sampling_period"] = sampling_period

                            self.logger.debug("")

                            continue  # skip memmap, chan info for time col

                        else:

                            raise NotImplementedError(
                                "series data are supported only for the first " "data column (time)!"
                            )

                    elif col_type == 4:

                        # short
                        dtype = "h"
                        gain, offset = 1, 0  # data neither scaled nor off-set

                    elif col_type == 5:

                        # long
                        dtype = "l"
                        gain, offset = 1, 0  # data neither scaled nor off-set

                    elif col_type == 6:

                        # float
                        dtype = "f"
                        gain, offset = 1, 0  # data neither scaled nor off-set

                    elif col_type == 7:

                        # double
                        dtype = "d"
                        gain, offset = 1, 0  # data neither scaled nor off-set

                    elif col_type == 10:

                        # scaled short
                        dtype = "h"
                        gain, offset = f.read_f("dd")  # data scaled w/ offset

                    else:

                        raise NotImplementedError(f'unimplemented column type "{col_type}"!')

                else:

                    raise NotImplementedError(f'unimplemented file format version "{format_ver}"!')

                ##############################################
                # COLUMN MEMMAP AND CHANNEL INFO

                # create a memory map that allows accessing parts of the file
                # without loading it all into memory
                array = np.memmap(self.filename, mode="r", dtype=f.byte_order + dtype, offset=f.tell(), shape=n_points)

                # advance the file position to after the data array
                f.seek(array.nbytes, 1)

                if i == 0:
                    # assume this is the time column containing n_points values

                    # verify times are spaced regularly
                    diffs = np.diff(array)
                    increment = np.median(diffs)
                    max_frac_step_deviation = np.max(np.abs(diffs / increment - 1))
                    tolerance = 1e-3
                    if max_frac_step_deviation > tolerance:
                        self.logger.debug(
                            "largest proportional deviation "
                            "from median step size in the first "
                            "column exceeds the tolerance "
                            "of " + str(tolerance) + ":"
                            " " + str(max_frac_step_deviation)
                        )
                        raise ValueError("first data column (assumed to be " "time) is not regularly spaced")

                    first_value = array[0]

                    self.logger.debug(f"interval: {increment}, freq: {1 / increment}")
                    self.logger.debug(f"start: {first_value}, end: {first_value + increment * (n_points - 1)}")

                    t_start, sampling_period = first_value, increment
                    self.info["t_start"] = t_start
                    self.info["sampling_period"] = sampling_period

                    self.logger.debug("")

                    continue  # skip saving memmap, chan info for time col

                else:
                    # not a time column

                    self.logger.debug(f"gain: {gain}, offset: {offset}")
                    self.logger.debug("initial data: {array[:5] * gain + offset}")

                    # channel_info will be cast to _signal_channel_dtype
                    channel_info = (name, str(i), 1 / sampling_period, f.byte_order + dtype, units, gain, offset, "0")

                    self.logger.debug("channel_info: {channel_info}")
                    self.logger.debug("")

                    sig_memmaps.append(array)
                    sig_channels.append(channel_info)

            # END COLUMNS
            ##############################################

            # initialize lists for events and epochs
            raw_event_timestamps = []
            raw_epoch_timestamps = []
            raw_epoch_durations = []
            event_labels = []
            epoch_labels = []

            # the remainder of the file may contain metadata, events and epochs
            try:

                ##############################################
                # COMMENT

                self.logger.debug("== COMMENT ==")

                comment = f.read_f("S")
                self.info["comment"] = comment

                self.logger.debug(comment if comment else "no comment!")
                self.logger.debug("")

                ##############################################
                # NOTES

                self.logger.debug("== NOTES ==")

                notes = f.read_f("S")
                self.info["notes"] = notes

                self.logger.debug(notes if notes else "no notes!")
                self.logger.debug("")

                ##############################################
                # TRACES

                self.logger.debug("== TRACES ==")

                n_traces = f.read_f("l")
                self.info["n_traces"] = n_traces

                self.logger.debug(f"n_traces: {n_traces}")
                self.logger.debug("")

                trace_header_info_list = {}
                group_ids = []
                for i in range(n_traces):

                    # AxoGraph traces are 1-indexed in GUI, so use i+1 below
                    self.logger.debug(f"== TRACE #{i+1} ==")

                    trace_header_info = {}

                    if format_ver < 6:
                        # before format version 6, there was only one version
                        # of the header, and version numbers were not provided
                        trace_header_info["trace_header_version"] = 1
                    else:
                        # for format versions 6 and later, the header version
                        # must be read
                        trace_header_info["trace_header_version"] = f.read_f("l")

                    if trace_header_info["trace_header_version"] == 1:
                        TraceHeaderDescription = TraceHeaderDescriptionV1
                    elif trace_header_info["trace_header_version"] == 2:
                        TraceHeaderDescription = TraceHeaderDescriptionV2
                    else:
                        raise NotImplementedError(
                            f'unimplemented trace header version `{trace_header_info["trace_header_version"]}`!'
                        )

                    for key, fmt in TraceHeaderDescription:
                        trace_header_info[key] = f.read_f(fmt)
                    # AxoGraph traces are 1-indexed in GUI, so use i+1 below
                    trace_header_info_list[i + 1] = trace_header_info
                    group_ids.append(trace_header_info["group_id_for_this_trace"])

                    self.logger.debug(trace_header_info)
                    self.logger.debug("")
                self.info["trace_header_info_list"] = trace_header_info_list

                ##############################################
                # GROUPS

                self.logger.debug("== GROUPS ==")

                n_groups = f.read_f("l")
                self.info["n_groups"] = n_groups
                group_ids = np.sort(list(set(group_ids)))  # remove duplicates and sort
                if n_groups != len(group_ids):
                    raise ValueError(f"expected group_ids to have length {n_groups}: {group_ids}")

                self.logger.debug(f"n_groups: {n_groups}")
                self.logger.debug(f"group_ids: {group_ids}")
                self.logger.debug("")

                group_header_info_list = {}
                for i in group_ids:

                    # AxoGraph groups are 0-indexed in GUI, so use i below
                    self.logger.debug(f"== GROUP #{i} ==")

                    group_header_info = {}

                    if format_ver < 6:
                        # before format version 6, there was only one version
                        # of the header, and version numbers were not provided
                        group_header_info["group_header_version"] = 1
                    else:
                        # for format versions 6 and later, the header version
                        # must be read
                        group_header_info["group_header_version"] = f.read_f("l")

                    if group_header_info["group_header_version"] == 1:
                        GroupHeaderDescription = GroupHeaderDescriptionV1
                    else:
                        raise NotImplementedError(
                            f'unimplemented group header version `{group_header_info["group_header_version"]}`!'
                        )

                    for key, fmt in GroupHeaderDescription:
                        group_header_info[key] = f.read_f(fmt)
                    # AxoGraph groups are 0-indexed in GUI, so use i below
                    group_header_info_list[i] = group_header_info

                    self.logger.debug(group_header_info)
                    self.logger.debug("")
                self.info["group_header_info_list"] = group_header_info_list

                ##############################################
                # UNKNOWN

                self.logger.debug(">> UNKNOWN 1 <<")

                # 36 bytes of undeciphered data (types here are guesses)
                unknowns = f.read_f("9l")

                self.logger.debug(unknowns)
                self.logger.debug("")

                ##############################################
                # EPISODES

                self.logger.debug("== EPISODES ==")

                # a subset of episodes can be selected for "review", or
                # episodes can be paged through one by one, and the indexes of
                # those currently in review appear in this list
                episodes_in_review = []
                n_episodes = f.read_f("l")
                self.info["n_episodes"] = n_episodes
                for i in range(n_episodes):
                    episode_bool = f.read_f("Z")
                    if episode_bool:
                        episodes_in_review.append(i + 1)
                self.info["episodes_in_review"] = episodes_in_review

                self.logger.debug(f"n_episodes: {n_episodes}")
                self.logger.debug(f"episodes_in_review: {episodes_in_review}")

                if format_ver == 5:

                    # the test file for version 5 contains this extra list of
                    # episode indexes with unknown purpose
                    old_unknown_episode_list = []
                    n_episodes2 = f.read_f("l")
                    for i in range(n_episodes2):
                        episode_bool = f.read_f("Z")
                        if episode_bool:
                            old_unknown_episode_list.append(i + 1)

                    self.logger.debug(f"old_unknown_episode_list: {old_unknown_episode_list}")
                    if n_episodes2 != n_episodes:
                        self.logger.debug(f"n_episodes2 ({n_episodes2}) and n_episodes ({n_episodes}) " "differ!")

                # another list of episode indexes with unknown purpose
                unknown_episode_list = []
                n_episodes3 = f.read_f("l")
                for i in range(n_episodes3):
                    episode_bool = f.read_f("Z")
                    if episode_bool:
                        unknown_episode_list.append(i + 1)

                self.logger.debug(f"unknown_episode_list: {unknown_episode_list}")
                if n_episodes3 != n_episodes:
                    self.logger.debug(f"n_episodes3 ({n_episodes3,}) and n_episodes ({n_episodes}) " "differ!")

                # episodes can be masked to be removed from the pool of
                # reviewable episodes completely until unmasked, and the
                # indexes of those currently masked appear in this list
                masked_episodes = []
                n_episodes4 = f.read_f("l")
                for i in range(n_episodes4):
                    episode_bool = f.read_f("Z")
                    if episode_bool:
                        masked_episodes.append(i + 1)
                self.info["masked_episodes"] = masked_episodes

                self.logger.debug(f"masked_episodes: {masked_episodes}")
                if n_episodes4 != n_episodes:
                    self.logger.debug(f"n_episodes4 ({n_episodes4}) and n_episodes ({n_episodes}) " "differ!")
                self.logger.debug("")

                ##############################################
                # UNKNOWN

                self.logger.debug(">> UNKNOWN 2 <<")

                # 68 bytes of undeciphered data (types here are guesses)
                unknowns = f.read_f("d 9l d 4l")

                self.logger.debug(unknowns)
                self.logger.debug("")

                ##############################################
                # FONTS

                if format_ver >= 6:
                    font_categories = ["axis titles", "axis labels (ticks)", "notes", "graph title"]
                else:
                    # would need an old version of AxoGraph to determine how it
                    # used these settings
                    font_categories = ["everything (?)"]

                font_settings_info_list = {}
                for i in font_categories:

                    self.logger.debug(f"== FONT SETTINGS FOR {i} ==")

                    font_settings_info = {}
                    for key, fmt in FontSettingsDescription:
                        font_settings_info[key] = f.read_f(fmt)

                    # I don't know why two arbitrary values were selected to
                    # represent this switch, but it seems they were
                    # - setting1 could contain other undeciphered data as a
                    #   bitmask, like setting2
                    if font_settings_info["setting1"] not in [
                        FONT_BOLD,
                        FONT_NOT_BOLD,
                    ]:
                        raise ValueError(
                            f"expected setting1 ({ font_settings_info['setting1']}) to have value FONT_BOLD "
                            f"({FONT_BOLD}) or FONT_NOT_BOLD ({FONT_NOT_BOLD})"
                        )

                    # size is stored 10 times bigger than real value
                    font_settings_info["size"] = font_settings_info["size"] / 10.0
                    font_settings_info["bold"] = bool(font_settings_info["setting1"] == FONT_BOLD)
                    font_settings_info["italics"] = bool(font_settings_info["setting2"] & FONT_ITALICS)
                    font_settings_info["underline"] = bool(font_settings_info["setting2"] & FONT_UNDERLINE)
                    font_settings_info["strikeout"] = bool(font_settings_info["setting2"] & FONT_STRIKEOUT)
                    font_settings_info_list[i] = font_settings_info

                    self.logger.debug(font_settings_info)
                    self.logger.debug("")
                self.info["font_settings_info_list"] = font_settings_info_list

                ##############################################
                # X-AXIS SETTINGS

                self.logger.debug("== X-AXIS SETTINGS ==")

                x_axis_settings_info = {}
                for key, fmt in XAxisSettingsDescription:
                    x_axis_settings_info[key] = f.read_f(fmt)
                self.info["x_axis_settings_info"] = x_axis_settings_info

                self.logger.debug(x_axis_settings_info)
                self.logger.debug("")

                ##############################################
                # UNKNOWN

                self.logger.debug(">> UNKNOWN 3 <<")

                # 108 bytes of undeciphered data (types here are guesses)
                unknowns = f.read_f("8l 3d 13l")

                self.logger.debug(unknowns)
                self.logger.debug("")

                ##############################################
                # EVENTS / TAGS

                self.logger.debug("=== EVENTS / TAGS ===")

                n_events, n_events_again = f.read_f("ll")
                self.info["n_events"] = n_events

                self.logger.debug(f"n_events: {n_events}")

                # event / tag timing is stored as an index into time
                raw_event_timestamps = []
                event_labels = []
                for i in range(n_events_again):
                    event_index = f.read_f("l")
                    raw_event_timestamps.append(event_index)
                n_events_yet_again = f.read_f("l")
                for i in range(n_events_yet_again):
                    title = f.read_f("S")
                    event_labels.append(title)

                event_list = []
                for event_label, event_index in zip(event_labels, raw_event_timestamps):
                    # t_start shouldn't be added here
                    event_time = event_index * sampling_period
                    event_list.append({"title": event_label, "index": event_index, "time": event_time})
                self.info["event_list"] = event_list
                for event in event_list:
                    self.logger.debug(event)
                self.logger.debug("")

                ##############################################
                # UNKNOWN

                self.logger.debug(">> UNKNOWN 4 <<")

                # 28 bytes of undeciphered data (types here are guesses)
                unknowns = f.read_f("7l")

                self.logger.debug(unknowns)
                self.logger.debug("")

                ##############################################
                # EPOCHS / INTERVAL BARS

                self.logger.debug("=== EPOCHS / INTERVAL BARS ===")

                n_epochs = f.read_f("l")
                self.info["n_epochs"] = n_epochs

                self.logger.debug(f"n_epochs: {n_epochs}")

                epoch_list = []
                for i in range(n_epochs):
                    epoch_info = {}
                    for key, fmt in EpochInfoDescription:
                        epoch_info[key] = f.read_f(fmt)
                    epoch_list.append(epoch_info)
                self.info["epoch_list"] = epoch_list

                # epoch / interval bar timing and duration are stored in
                # seconds, so here they are converted to (possibly non-integer)
                # indexes into time to fit into the procrustean beds of
                # _rescale_event_timestamp and _rescale_epoch_duration
                raw_epoch_timestamps = []
                raw_epoch_durations = []
                epoch_labels = []
                for epoch in epoch_list:
                    raw_epoch_timestamps.append(epoch["t_start"] / sampling_period)
                    raw_epoch_durations.append((epoch["t_stop"] - epoch["t_start"]) / sampling_period)
                    epoch_labels.append(epoch["title"])
                    self.logger.debug(epoch)
                self.logger.debug("")

                ##############################################
                # UNKNOWN

                self.logger.debug(">> UNKNOWN 5 (includes y-axis plot ranges) <<")

                # lots of undeciphered data
                rest_of_the_file = f.read()

                self.logger.debug(rest_of_the_file)
                self.logger.debug("")

                self.logger.debug("End of file reached (expected)")

            except EOFError as e:
                if format_ver == 1 or format_ver == 2:
                    # for format versions 1 and 2, metadata like graph display
                    # information was stored separately in the "resource fork"
                    # of the file, so reaching the end of the file before all
                    # metadata is parsed is expected
                    self.logger.debug("End of file reached (expected)")
                    pass
                else:
                    # for format versions 3 and later, there should be metadata
                    # stored at the end of the file, so warn that something may
                    # have gone wrong, but try to continue anyway
                    self.logger.warning(
                        "End of file reached unexpectedly " "while parsing metadata, will attempt " "to continue"
                    )
                    self.logger.debug(e, exc_info=True)
                    pass

            except UnicodeDecodeError as e:
                # warn that something went wrong with reading a string, but try
                # to continue anyway
                self.logger.warning(
                    "Problem decoding text while parsing "
                    "metadata, will ignore any remaining "
                    "metadata and attempt to continue"
                )
                self.logger.debug(e, exc_info=True)
                pass

        self.logger.debug("")

        ##############################################
        # RAWIO HEADER

        # event_channels will be cast to _event_channel_dtype
        event_channels = []
        event_channels.append(("AxoGraph Tags", "", "event"))
        event_channels.append(("AxoGraph Intervals", "", "epoch"))

        if len(sig_channels) > 0:
            signal_streams = [("Signals", "0")]
        else:
            signal_streams = []

        # organize header
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = np.array(signal_streams, dtype=_signal_stream_dtype)
        self.header["signal_channels"] = np.array(sig_channels, dtype=_signal_channel_dtype)
        self.header["event_channels"] = np.array(event_channels, dtype=_event_channel_dtype)
        self.header["spike_channels"] = np.array([], dtype=_spike_channel_dtype)

        ##############################################
        # DATA OBJECTS

        # organize data
        self._sampling_period = sampling_period
        self._t_start = t_start
        self._raw_signals = [sig_memmaps]  # first index is seg_index
        self._raw_event_epoch_timestamps = [np.array(raw_event_timestamps), np.array(raw_epoch_timestamps)]
        self._raw_event_epoch_durations = [None, np.array(raw_epoch_durations)]
        self._event_epoch_labels = [np.array(event_labels, dtype="U"), np.array(epoch_labels, dtype="U")]


class StructFile(BufferedReader):
    """
    A container for the file buffer with some added convenience functions for
    reading AxoGraph files
    """

    def __init__(self, *args, **kwargs):
        # As far as I've seen, every AxoGraph file uses big-endian encoding,
        # regardless of the system architecture on which it was created, but
        # here I provide means for controlling byte ordering in case a counter
        # example is found.
        self.byte_order = kwargs.pop("byte_order", ">")
        if self.byte_order == ">":
            # big-endian
            self.utf_16_decoder = "utf-16-be"
        elif self.byte_order == "<":
            # little-endian
            self.utf_16_decoder = "utf-16-le"
        else:
            # unspecified
            self.utf_16_decoder = "utf-16"
        super().__init__(*args, **kwargs)

    def read_and_unpack(self, fmt):
        """
        Calculate the number of bytes corresponding to the format string, read
        in that number of bytes, and unpack them according to the format string
        """
        try:
            return unpack(self.byte_order + fmt, self.read(calcsize(self.byte_order + fmt)))
        except Exception as e:
            if e.args[0].startswith("unpack requires a buffer of"):
                raise EOFError(e)
            else:
                raise

    def read_string(self):
        """
        The most common string format in AxoGraph files is a variable length
        string with UTF-16 encoding, preceded by a 4-byte integer (long)
        specifying the length of the string in bytes. Unlike a Pascal string
        ('p' format), these strings are not stored in a fixed number of bytes
        with padding at the end. This function reads in one of these variable
        length strings
        """

        # length may be -1, 0, or a positive integer
        length = self.read_and_unpack("l")[0]
        if length > 0:
            return self.read(length).decode(self.utf_16_decoder)
        else:
            return ""

    def read_bool(self):
        """
        AxoGraph files encode each boolean as 4-byte integer (long) with value
        1 = True, 0 = False. This function reads in one of these booleans.
        """
        return bool(self.read_and_unpack("l")[0])

    def read_f(self, fmt, offset=None):
        """
        This function is a wrapper for read_and_unpack that adds compatibility
        with two new format strings:
            'S': a variable length UTF-16 string, readable with read_string
            'Z': a boolean encoded as a 4-byte integer, readable with read_bool
        This method does not implement support for numbers before the new
        format strings, such as '3Z' to represent 3 bools (use 'ZZZ' instead).
        """

        if offset is not None:
            self.seek(offset)

        # place commas before and after each instance of S or Z
        for special in ["S", "Z"]:
            fmt = fmt.replace(special, "," + special + ",")

        # split S and Z into isolated strings
        fmt = fmt.split(",")

        # construct a tuple of unpacked data
        data = ()
        for subfmt in fmt:
            if subfmt == "S":
                data += (self.read_string(),)
            elif subfmt == "Z":
                data += (self.read_bool(),)
            else:
                data += self.read_and_unpack(subfmt)

        if len(data) == 1:
            return data[0]
        else:
            return data


FONT_BOLD = 75  # mysterious arbitrary constant
FONT_NOT_BOLD = 50  # mysterious arbitrary constant
FONT_ITALICS = 1
FONT_UNDERLINE = 2
FONT_STRIKEOUT = 4

TraceHeaderDescriptionV1 = [
    # documented in AxoGraph_ReadWrite.h
    ("x_index", "l"),
    ("y_index", "l"),
    ("err_bar_index", "l"),
    ("group_id_for_this_trace", "l"),
    ("hidden", "Z"),  # AxoGraph_ReadWrite.h incorrectly states "shown" instead
    ("min_x", "d"),
    ("max_x", "d"),
    ("min_positive_x", "d"),
    ("x_is_regularly_spaced", "Z"),
    ("x_increases_monotonically", "Z"),
    ("x_interval_if_regularly_spaced", "d"),
    ("min_y", "d"),
    ("max_y", "d"),
    ("min_positive_y", "d"),
    ("trace_color", "xBBB"),
    ("display_joined_line_plot", "Z"),
    ("line_thickness", "d"),
    ("pen_style", "l"),
    ("display_symbol_plot", "Z"),
    ("symbol_type", "l"),
    ("symbol_size", "l"),
    ("draw_every_data_point", "Z"),
    ("skip_points_by_distance_instead_of_pixels", "Z"),
    ("pixels_between_symbols", "l"),
    ("display_histogram_plot", "Z"),
    ("histogram_type", "l"),
    ("histogram_bar_separation", "l"),
    ("display_error_bars", "Z"),
    ("display_pos_err_bar", "Z"),
    ("display_neg_err_bar", "Z"),
    ("err_bar_width", "l"),
]

# documented in AxoGraph_ReadWrite.h
# - only one difference exists between versions 1 and 2
TraceHeaderDescriptionV2 = list(TraceHeaderDescriptionV1)  # make a copy
TraceHeaderDescriptionV2.insert(3, ("neg_err_bar_index", "l"))

GroupHeaderDescriptionV1 = [
    # undocumented and reverse engineered
    ("title", "S"),
    ("unknown1", "h"),  # 2 bytes of undeciphered data (types are guesses)
    ("units", "S"),
    ("unknown2", "hll"),  # 10 bytes of undeciphered data (types are guesses)
]

FontSettingsDescription = [
    # undocumented and reverse engineered
    ("font", "S"),
    ("size", "h"),  # divide this 2-byte integer by 10 to get font size
    ("unknown1", "5b"),  # 5 bytes of undeciphered data (types are guesses)
    ("setting1", "B"),  # includes bold setting
    ("setting2", "B"),  # italics, underline, strikeout specified in bitmap
]

XAxisSettingsDescription = [
    # undocumented and reverse engineered
    ("unknown1", "3l2d"),  # 28 bytes of undeciphered data (types are guesses)
    ("plotted_x_range", "dd"),
    ("unknown2", "d"),  # 8 bytes of undeciphered data (types are guesses)
    ("auto_x_ticks", "Z"),
    ("x_minor_ticks", "d"),
    ("x_major_ticks", "d"),
    ("x_axis_title", "S"),
    ("unknown3", "h"),  # 2 bytes of undeciphered data (types are guesses)
    ("units", "S"),
    ("unknown4", "h"),  # 2 bytes of undeciphered data (types are guesses)
]

EpochInfoDescription = [
    # undocumented and reverse engineered
    ("title", "S"),
    ("t_start", "d"),
    ("t_stop", "d"),
    ("y_pos", "d"),
]



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.tdtrawio import TdtRawIO


class TdtIO(TdtRawIO, BaseFromRaw):
    """
    Class for reading data from from Tucker Davis TTank format.

    Terminology:
    TDT holds data with tanks (actually a directory). And tanks hold sub blocks
    (sub directories).
    Tanks correspond to Neo Blocks and TDT blocks correspond to Neo Segments.
    """

    _prefered_signal_group_mode = "group-by-same-units"
    mode = "dir"

    def __init__(self, dirname):
        TdtRawIO.__init__(self, dirname=dirname)
        BaseFromRaw.__init__(self, dirname)



"""
Class for reading data in CED spike2 files (.smr).

This code is based on:
 - sonpy, written by Antonio Gonzalez <Antonio.Gonzalez@cantab.net>
    Available here ::
    http://www.neuro.ki.se/broberger/

and sonpy come from :
 - SON Library 2.0 for MATLAB, written by Malcolm Lidierth at
    King's College London.
    See http://www.kcl.ac.uk/depsta/biomedical/cfnr/lidierth.html

This IO support old (<v6) and new files (>v7) of spike2


Author: Samuel Garcia

"""

from collections import OrderedDict

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class Spike2RawIO(BaseRawIO):
    """
    This implementation in neo reads only old smr files.
    For smrx files you need to use CedRawIO which is based on sonpy.

    Parameters
    ----------
    filename: str, default: ''
        The *.smr file to be loaded
    take_ideal_sampling_rate: bool, default: False
        If True takes the `ideal_rate` from info
    ced_units: bool, default: True
        If True uses the ced unit ids
    try_signal_grouping: bool, default: True
        If True will attempt to group signals together

    """

    extensions = ["smr"]
    rawmode = "one-file"

    def __init__(self, filename="", take_ideal_sampling_rate=False, ced_units=True, try_signal_grouping=True):
        BaseRawIO.__init__(self)
        self.filename = filename

        self.take_ideal_sampling_rate = take_ideal_sampling_rate
        self.ced_units = ced_units
        self.try_signal_grouping = try_signal_grouping

    def _parse_header(self):

        # get header info and channel_info
        with open(self.filename, "rb") as fid:
            self._global_info = read_as_dict(fid, headerDescription)
            info = self._global_info
            if info["system_id"] < 6:
                info["dtime_base"] = 1e-6
                info["datetime_detail"] = 0
                info["datetime_year"] = 0
            if info["system_id"] == 9:
                self._diskblock = DISKBLOCK
            else:
                self._diskblock = 1

            self._time_factor = info["us_per_time"] * info["dtime_base"]

            self._channel_infos = []
            for chan_id in range(info["channels"]):
                fid.seek(512 + 140 * chan_id)
                chan_info = read_as_dict(fid, channelHeaderDesciption1)

                if chan_info["kind"] in [1, 6]:
                    dt = [
                        ("scale", "f4"),
                        ("offset", "f4"),
                        ("unit", "S6"),
                    ]
                    chan_info.update(read_as_dict(fid, dt))

                elif chan_info["kind"] in [7, 9]:
                    dt = [
                        ("min", "f4"),
                        ("max", "f4"),
                        ("unit", "S6"),
                    ]
                    chan_info.update(read_as_dict(fid, dt))

                elif chan_info["kind"] in [4]:
                    dt = [
                        ("init_low", "u1"),
                        ("next_low", "u1"),
                    ]
                    chan_info.update(read_as_dict(fid, dt))

                if chan_info["kind"] in [1, 6, 7, 9]:
                    if info["system_id"] < 6:
                        chan_info.update(read_as_dict(fid, [("divide", "i2")]))
                    else:
                        chan_info.update(read_as_dict(fid, [("interleave", "i2")]))

                chan_info["type"] = dict_kind[chan_info["kind"]]

                if chan_info["blocks"] == 0:
                    chan_info["t_start"] = 0.0  # this means empty signals
                else:
                    fid.seek(chan_info["firstblock"] * self._diskblock)
                    block_info = read_as_dict(fid, blockHeaderDesciption)
                    chan_info["t_start"] = (
                        float(block_info["start_time"]) * float(info["us_per_time"]) * float(info["dtime_base"])
                    )

                self._channel_infos.append(chan_info)

        # get data blocks index for all channel
        # run through all data block of of channel to prepare chan to block maps
        self._memmap = np.memmap(self.filename, dtype="u1", offset=0, mode="r")
        self._all_data_blocks = {}
        self._by_seg_data_blocks = {}
        for chan_id, chan_info in enumerate(self._channel_infos):
            data_blocks = []
            ind = chan_info["firstblock"] * self._diskblock
            for b in range(chan_info["blocks"]):
                block_info = self._memmap[ind : ind + 20].view(blockHeaderDesciption)[0]
                data_blocks.append((ind, block_info["items"], 0, block_info["start_time"], block_info["end_time"]))
                ind = block_info["succ_block"] * self._diskblock

            data_blocks = np.array(
                data_blocks,
                dtype=[
                    ("pos", "int32"),
                    ("size", "int32"),
                    ("cumsum", "int32"),
                    ("start_time", "int32"),
                    ("end_time", "int32"),
                ],
            )
            data_blocks["pos"] += 20  # 20 is the header size

            self._all_data_blocks[chan_id] = data_blocks
            self._by_seg_data_blocks[chan_id] = []

        # For all signal channel detect gaps between data block (pause in rec) so new Segment.
        # then check that all channel have the same gaps.
        # this part is tricky because we need to check that all channel have same pause.
        all_gaps_block_ind = {}
        for chan_id, chan_info in enumerate(self._channel_infos):
            if chan_info["kind"] in [1, 9]:
                data_blocks = self._all_data_blocks[chan_id]
                sig_size = np.sum(self._all_data_blocks[chan_id]["size"])
                if sig_size > 0:
                    interval = int(np.round(get_sample_interval(info, chan_info) / self._time_factor))
                    # detect gaps
                    inter_block_sizes = data_blocks["start_time"][1:] - data_blocks["end_time"][:-1]
                    (gaps_block_ind,) = np.nonzero(inter_block_sizes > interval)
                    all_gaps_block_ind[chan_id] = gaps_block_ind

        # find t_start/t_stop for each seg based on gaps index
        self._sig_t_starts = {}
        self._sig_t_stops = {}
        if len(all_gaps_block_ind) == 0:
            # this means no signal channels
            nb_segment = 1
            # loop over event/spike channel to get the min/max time
            t_start, t_stop = None, None
            for chan_id, chan_info in enumerate(self._channel_infos):
                data_blocks = self._all_data_blocks[chan_id]
                if data_blocks.size > 0:
                    # if t_start is None or data_blocks[0]['start_time']<t_start:
                    # t_start = data_blocks[0]['start_time']
                    if t_stop is None or data_blocks[-1]["end_time"] > t_stop:
                        t_stop = data_blocks[-1]["end_time"]
            # self._seg_t_starts = [t_start]
            self._seg_t_starts = [0]
            self._seg_t_stops = [t_stop]
        else:
            all_nb_seg = np.array([v.size + 1 for v in all_gaps_block_ind.values()])
            if np.all(all_nb_seg[0] != all_nb_seg):
                raise NeoReadWriteError("Signal channels have different pauses so different nb_segment")
            nb_segment = int(all_nb_seg[0])

            for chan_id, gaps_block_ind in all_gaps_block_ind.items():
                data_blocks = self._all_data_blocks[chan_id]
                self._sig_t_starts[chan_id] = []
                self._sig_t_stops[chan_id] = []

                for seg_ind in range(nb_segment):
                    if seg_ind == 0:
                        fisrt_bl = 0
                    else:
                        fisrt_bl = gaps_block_ind[seg_ind - 1] + 1
                    self._sig_t_starts[chan_id].append(data_blocks[fisrt_bl]["start_time"])

                    if seg_ind < nb_segment - 1:
                        last_bl = gaps_block_ind[seg_ind]
                    else:
                        last_bl = data_blocks.size - 1

                    self._sig_t_stops[chan_id].append(data_blocks[last_bl]["end_time"])

                    in_seg_data_block = data_blocks[fisrt_bl : last_bl + 1]
                    in_seg_data_block["cumsum"][1:] = np.cumsum(in_seg_data_block["size"][:-1])
                    self._by_seg_data_blocks[chan_id].append(in_seg_data_block)

            self._seg_t_starts = []
            self._seg_t_stops = []
            for seg_ind in range(nb_segment):
                # there is a small delay between all channel so take the max/min for t_start/t_stop
                t_start = min(self._sig_t_starts[chan_id][seg_ind] for chan_id in self._sig_t_starts)
                t_stop = max(self._sig_t_stops[chan_id][seg_ind] for chan_id in self._sig_t_stops)
                self._seg_t_starts.append(t_start)
                self._seg_t_stops.append(t_stop)

        # create typed channels
        signal_channels = []
        spike_channels = []
        event_channels = []

        self.internal_unit_ids = {}
        for chan_id, chan_info in enumerate(self._channel_infos):
            if chan_info["kind"] in [1, 6, 7, 9]:
                if self.take_ideal_sampling_rate:
                    sampling_rate = info["ideal_rate"]
                else:
                    sample_interval = get_sample_interval(info, chan_info)
                    sampling_rate = 1.0 / sample_interval

            name = chan_info["title"]

            if chan_info["kind"] in [1, 9]:
                # AnalogSignal
                if chan_id not in self._sig_t_starts:
                    continue
                units = chan_info["unit"]
                if chan_info["kind"] == 1:  # int16
                    gain = chan_info["scale"] / 6553.6
                    offset = chan_info["offset"]
                    sig_dtype = "int16"
                elif chan_info["kind"] == 9:  # float32
                    gain = 1.0
                    offset = 0.0
                    sig_dtype = "float32"
                stream_id = "0"  # set it after the loop
                signal_channels.append((name, str(chan_id), sampling_rate, sig_dtype, units, gain, offset, stream_id))

            elif chan_info["kind"] in [2, 3, 4, 5, 8]:
                # Event
                event_channels.append((name, chan_id, "event"))

            elif chan_info["kind"] in [6, 7]:  # SpikeTrain with waveforms
                wf_units = chan_info["unit"]
                if chan_info["kind"] == 6:
                    wf_gain = chan_info["scale"] / 6553.6
                    wf_offset = chan_info["offset"]
                    wf_left_sweep = chan_info["n_extra"] // 4
                elif chan_info["kind"] == 7:
                    wf_gain = 1.0
                    wf_offset = 0.0
                    wf_left_sweep = chan_info["n_extra"] // 8
                wf_sampling_rate = sampling_rate
                if self.ced_units:
                    # this is a huge pain because need
                    # to jump over all blocks
                    data_blocks = self._all_data_blocks[chan_id]
                    dt = get_channel_dtype(chan_info)
                    unit_ids = set()
                    for bl in range(data_blocks.size):
                        ind0 = data_blocks[bl]["pos"]
                        ind1 = data_blocks[bl]["size"] * dt.itemsize + ind0
                        raw_data = self._memmap[ind0:ind1].view(dt)
                        marker = raw_data["marker"] & 255
                        unit_ids.update(np.unique(marker))
                    unit_ids = sorted(list(unit_ids))
                else:
                    # All spike from one channel are group in one SpikeTrain
                    unit_ids = ["all"]
                for unit_id in unit_ids:
                    unit_index = len(spike_channels)
                    self.internal_unit_ids[unit_index] = (chan_id, unit_id)
                    _id = f"ch{chan_id}#{unit_id}"
                    spike_channels.append((name, _id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        if len(signal_channels) > 0:
            if self.try_signal_grouping:
                # try to group signals channel if same sampling_rate/dtype/...
                # it can raise error for some files (when they do not have signal length)
                common_keys = ["sampling_rate", "dtype", "units", "gain", "offset"]
                characteristics = signal_channels[common_keys]
                unique_characteristics = np.unique(characteristics)
                self._sig_dtypes = {}
                signal_streams = []
                for stream_index, charact in enumerate(unique_characteristics):
                    (chan_grp_indexes,) = np.nonzero(characteristics == charact)
                    stream_id = str(stream_index)
                    signal_channels["stream_id"][chan_grp_indexes] = stream_id

                    # check same size for channel in groups
                    for seg_index in range(nb_segment):
                        sig_sizes = []
                        for ind in chan_grp_indexes:
                            chan_id = int(signal_channels[ind]["id"])
                            sig_size = np.sum(self._by_seg_data_blocks[chan_id][seg_index]["size"])
                            sig_sizes.append(sig_size)
                        sig_sizes = np.array(sig_sizes)
                        if np.all(sig_sizes != sig_sizes[0]):
                            raise NeoReadWriteError(
                                "Signal channel in groups do not have same size, use try_signal_grouping=False"
                            )
                    self._sig_dtypes[stream_id] = np.dtype(charact["dtype"])
                    signal_streams.append((f"Signal stream {stream_id}", stream_id))
                signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
            else:
                # if try_signal_grouping fail the user can try to split each channel in
                # separate group
                signal_channels["stream_id"] = np.arange(signal_channels.size)
                signal_streams = np.zeros(signal_channels.size, dtype=_signal_stream_dtype)
                signal_streams["id"] = signal_channels["stream_id"]
                signal_streams["name"] = signal_channels["name"]
                self._sig_dtypes = {s["stream_id"]: np.dtype(s["dtype"]) for s in signal_channels}
        else:
            signal_streams = np.array([], dtype=_signal_stream_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotations
        self._generate_minimal_annotations()
        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["system_id"] = info["system_id"]
        for seg_index in range(nb_segment):
            seg_ann = self.raw_annotations["blocks"][0]["segments"][seg_index]
            seg_ann["system_id"] = info["system_id"]

            for c, stream_channel in enumerate(signal_streams):
                stream_id = stream_channel["id"]
                signal_an = self.raw_annotations["blocks"][0]["segments"][seg_index]["signals"][c]
                mask = signal_channels["stream_id"] == stream_id

                for key in ("phy_chan", "comment"):
                    values = []
                    for chan_id in signal_channels[mask]["id"]:
                        values.append(self._channel_infos[int(chan_id)][key])
                    signal_an["__array_annotations__"][key] = np.array(values)

            for c, unit_channel in enumerate(spike_channels):
                chan_id, unit_id = self.internal_unit_ids[c]
                unit_an = self.raw_annotations["blocks"][0]["segments"][seg_index]["spikes"][c]
                unit_an["physical_channel_index"] = self._channel_infos[chan_id]["phy_chan"]
                unit_an["comment"] = self._channel_infos[chan_id]["comment"]

            for c, event_channel in enumerate(event_channels):
                chan_id = int(event_channel["id"])
                ev_an = self.raw_annotations["blocks"][0]["segments"][seg_index]["events"][c]
                ev_an["physical_channel_index"] = self._channel_infos[chan_id]["phy_chan"]
                ev_an["comment"] = self._channel_infos[chan_id]["comment"]

    def _source_name(self):
        return self.filename

    def _segment_t_start(self, block_index, seg_index):
        return self._seg_t_starts[seg_index] * self._time_factor

    def _segment_t_stop(self, block_index, seg_index):
        return self._seg_t_stops[seg_index] * self._time_factor

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        chan_ids = signal_channels["id"]
        chan_id0 = int(chan_ids[0])
        sig_size = np.sum(self._by_seg_data_blocks[chan_id0][seg_index]["size"])
        return sig_size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        chan_ids = signal_channels["id"]
        chan_id0 = int(chan_ids[0])
        return self._sig_t_starts[chan_id0][seg_index] * self._time_factor

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._get_signal_size(block_index, seg_index, stream_index)

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        chan_ids = signal_channels["id"]
        self._sig_dtypes[stream_id]

        if channel_indexes is not None:
            chan_ids = chan_ids[channel_indexes]

        dt = self._sig_dtypes[stream_id]

        raw_signals = np.zeros((i_stop - i_start, len(chan_ids)), dtype=dt)
        for c, chan_id in enumerate(chan_ids):
            chan_id = int(chan_id)
            # NOTE: this actual way is slow because we run through
            # the file for each channel. The loop should be reversed.
            # But there is no guarantee that channels shared the same data block
            # indexes. So this make the job too difficult.
            data_blocks = self._by_seg_data_blocks[chan_id][seg_index]

            # loop over data blocks and get chunks
            bl0 = np.searchsorted(data_blocks["cumsum"], i_start, side="right") - 1
            bl1 = np.searchsorted(data_blocks["cumsum"], i_stop, side="right")
            ind = 0
            for bl in range(bl0, bl1):
                ind0 = data_blocks[bl]["pos"]
                ind1 = data_blocks[bl]["size"] * dt.itemsize + ind0
                data = self._memmap[ind0:ind1].view(dt)
                if bl == bl1 - 1:
                    # right border
                    # be careful that bl could be both bl0 and bl1!!
                    border = data.size - (i_stop - data_blocks[bl]["cumsum"])
                    if border > 0:
                        data = data[:-border]
                if bl == bl0:
                    # left border
                    border = i_start - data_blocks[bl]["cumsum"]
                    data = data[border:]
                raw_signals[ind : data.size + ind, c] = data
                ind += data.size
        return raw_signals

    def _count_in_time_slice(self, seg_index, chan_id, lim0, lim1, marker_filter=None):
        # count event or spike in time slice
        data_blocks = self._all_data_blocks[chan_id]
        chan_info = self._channel_infos[chan_id]
        dt = get_channel_dtype(chan_info)
        nb = 0
        for bl in range(data_blocks.size):
            ind0 = data_blocks[bl]["pos"]
            ind1 = data_blocks[bl]["size"] * dt.itemsize + ind0
            raw_data = self._memmap[ind0:ind1].view(dt)
            ts = raw_data["tick"]
            keep = (ts >= lim0) & (ts <= lim1)
            if marker_filter is not None:
                keep2 = (raw_data["marker"] & 255) == marker_filter
                keep = keep & keep2
            nb += np.sum(keep)
            if ts[-1] > lim1:
                break
        return nb

    def _get_internal_timestamp_(self, seg_index, chan_id, t_start, t_stop, other_field=None, marker_filter=None):
        chan_info = self._channel_infos[chan_id]
        # data_blocks = self._by_seg_data_blocks[chan_id][seg_index]
        data_blocks = self._all_data_blocks[chan_id]
        dt = get_channel_dtype(chan_info)

        if t_start is None:
            # lim0 = 0
            lim0 = self._seg_t_starts[seg_index]
        else:
            lim0 = int(t_start / self._time_factor)

        if t_stop is None:
            # lim1 = 2**32
            lim1 = self._seg_t_stops[seg_index]
        else:
            lim1 = int(t_stop / self._time_factor)

        timestamps = []
        othervalues = []
        for bl in range(data_blocks.size):
            ind0 = data_blocks[bl]["pos"]
            ind1 = data_blocks[bl]["size"] * dt.itemsize + ind0
            raw_data = self._memmap[ind0:ind1].view(dt)
            ts = raw_data["tick"]
            keep = (ts >= lim0) & (ts <= lim1)
            if marker_filter is not None:
                keep2 = (raw_data["marker"] & 255) == marker_filter
                keep = keep & keep2

            timestamps.append(ts[keep])
            if other_field is not None:
                othervalues.append(raw_data[other_field][keep])
            if ts[-1] > lim1:
                break

        if len(timestamps) > 0:
            timestamps = np.concatenate(timestamps)
        else:
            timestamps = np.zeros(0, dtype="int16")

        if other_field is None:
            return timestamps
        else:
            if len(timestamps) > 0:
                othervalues = np.concatenate(othervalues)
            else:
                othervalues = np.zeros(0, dtype=dt.fields[other_field][0])
            return timestamps, othervalues

    def _spike_count(self, block_index, seg_index, unit_index):
        chan_id, unit_id = self.internal_unit_ids[unit_index]
        if self.ced_units:
            marker_filter = unit_id
        else:
            marker_filter = None
        lim0 = self._seg_t_starts[seg_index]
        lim1 = self._seg_t_stops[seg_index]
        return self._count_in_time_slice(seg_index, chan_id, lim0, lim1, marker_filter=marker_filter)

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        unit_header = self.header["spike_channels"][unit_index]
        chan_id, unit_id = self.internal_unit_ids[unit_index]

        if self.ced_units:
            marker_filter = unit_id
        else:
            marker_filter = None

        spike_timestamps = self._get_internal_timestamp_(
            seg_index, chan_id, t_start, t_stop, marker_filter=marker_filter
        )

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times *= self._time_factor
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        unit_header = self.header["spike_channels"][unit_index]
        chan_id, unit_id = self.internal_unit_ids[unit_index]

        if self.ced_units:
            marker_filter = unit_id
        else:
            marker_filter = None

        timestamps, waveforms = self._get_internal_timestamp_(
            seg_index, chan_id, t_start, t_stop, other_field="waveform", marker_filter=marker_filter
        )

        waveforms = waveforms.reshape(timestamps.size, 1, -1)

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        event_header = self.header["event_channels"][event_channel_index]
        chan_id = int(event_header["id"])  # because set to string in header
        lim0 = self._seg_t_starts[seg_index]
        lim1 = self._seg_t_stops[seg_index]
        return self._count_in_time_slice(seg_index, chan_id, lim0, lim1, marker_filter=None)

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        event_header = self.header["event_channels"][event_channel_index]
        chan_id = int(event_header["id"])  # because set to string in header
        chan_info = self._channel_infos[chan_id]

        if chan_info["kind"] == 5:
            timestamps, labels = self._get_internal_timestamp_(
                seg_index, chan_id, t_start, t_stop, other_field="marker"
            )
            # the real encoding is unknown ASCII make it safe
            if labels.dtype.kind == "S":
                labels = np.char.decode(labels, "ascii", errors="ignore")
            else:
                labels = labels.astype("U")
        elif chan_info["kind"] == 8:
            timestamps, labels = self._get_internal_timestamp_(seg_index, chan_id, t_start, t_stop, other_field="label")
            # the real encoding is unknown ASCII make it safe
            if labels.dtype.kind == "S":
                labels = np.char.decode(labels, "ascii", errors="ignore")
            else:
                labels = labels.astype("U")
        else:
            timestamps = self._get_internal_timestamp_(seg_index, chan_id, t_start, t_stop, other_field=None)
            labels = np.zeros(timestamps.size, dtype="U")

        labels = labels.astype("U")
        durations = None

        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype)
        event_times *= self._time_factor
        return event_times


def read_as_dict(fid, dtype):
    """
    Given a file descriptor (seek at the good place externally)
    and a numpy.dtype of the binary struct return a dict.
    Make conversion for strings.
    """
    dt = np.dtype(dtype)
    h = np.frombuffer(fid.read(dt.itemsize), dt)[0]
    info = OrderedDict()
    for k in dt.names:
        v = h[k]

        if dt[k].kind == "S":
            v = v.decode("iso-8859-1")
            if len(v) > 0:
                length = ord(v[0])
                v = v[1 : length + 1]

        info[k] = v
    return info


def get_channel_dtype(chan_info):
    """
    Get dtype by kind.
    """
    if chan_info["kind"] == 1:  # Raw signal
        dt = "int16"
    elif chan_info["kind"] in [2, 3, 4]:  # Event data
        dt = [("tick", "i4")]
    elif chan_info["kind"] in [5]:  # Marker data
        dt = [("tick", "i4"), ("marker", "i4")]
    elif chan_info["kind"] in [6]:  # AdcMark data (waveform)
        dt = [
            ("tick", "i4"),
            ("marker", "i4"),
            # ('adc', 'S%d' % chan_info['n_extra'])]
            ("waveform", "int16", chan_info["n_extra"] // 2),
        ]
    elif chan_info["kind"] in [7]:  # RealMark data (waveform)
        dt = [
            ("tick", "i4"),
            ("marker", "i4"),
            # ('real', 'S%d' % chan_info['n_extra'])]
            ("waveform", "float32", chan_info["n_extra"] // 4),
        ]
    elif chan_info["kind"] in [8]:  # TextMark data
        dt = [("tick", "i4"), ("marker", "i4"), ("label", f"S{chan_info['n_extra']}")]
    elif chan_info["kind"] == 9:  # Float signal
        dt = "float32"
    dt = np.dtype(dt)
    return dt


def get_sample_interval(info, chan_info):
    """
    Get sample interval for one channel
    """
    if info["system_id"] in [1, 2, 3, 4, 5]:  # Before version 5
        sample_interval = (int(chan_info["divide"]) * info["us_per_time"] * info["time_per_adc"]) * 1e-6
    else:
        sample_interval = int(chan_info["l_chan_dvd"]) * info["us_per_time"] * info["dtime_base"]
    return sample_interval


# headers structures :
headerDescription = [
    ("system_id", "i2"),
    ("copyright", "S10"),
    ("creator", "S8"),
    ("us_per_time", "i2"),
    ("time_per_adc", "i2"),
    ("filestate", "i2"),
    ("first_data", "i4"),  # i8
    ("channels", "i2"),
    ("chan_size", "i2"),
    ("extra_data", "i2"),
    ("buffersize", "i2"),
    ("os_format", "i2"),
    ("max_ftime", "i4"),  # i8
    ("dtime_base", "f8"),
    ("datetime_detail", "u1"),
    ("datetime_year", "i2"),
    ("pad", "S52"),
    ("comment1", "S80"),
    ("comment2", "S80"),
    ("comment3", "S80"),
    ("comment4", "S80"),
    ("comment5", "S80"),
]

channelHeaderDesciption1 = [
    ("del_size", "i2"),
    ("next_del_block", "i4"),  # i8
    ("firstblock", "i4"),  # i8
    ("lastblock", "i4"),  # i8
    ("blocks", "i2"),
    ("n_extra", "i2"),
    ("pre_trig", "i2"),
    ("free0", "i2"),
    ("py_sz", "i2"),
    ("max_data", "i2"),
    ("comment", "S72"),
    ("max_chan_time", "i4"),  # i8
    ("l_chan_dvd", "i4"),  # i8
    ("phy_chan", "i2"),
    ("title", "S10"),
    ("ideal_rate", "f4"),
    ("kind", "u1"),
    ("unused1", "i1"),
]

blockHeaderDesciption = [
    ("pred_block", "i4"),  # i8
    ("succ_block", "i4"),  # i8
    ("start_time", "i4"),  # i8
    ("end_time", "i4"),  # i8
    ("channel_num", "i2"),
    ("items", "i2"),
]

dict_kind = {
    0: "empty",
    1: "Adc",
    2: "EventFall",
    3: "EventRise",
    4: "EventBoth",
    5: "Marker",
    6: "AdcMark",
    7: "RealMark",
    8: "TextMark",
    9: "RealWave",
}

DISKBLOCK = 512



"""
Tests of neo.io.asciispiketrainio
"""

import unittest

from neo.io import AsciiSpikeTrainIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestAsciiSpikeTrainIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = AsciiSpikeTrainIO
    entities_to_download = ["asciispiketrain"]
    entities_to_test = [
        "asciispiketrain/File_ascii_spiketrain_1.txt",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Class for reading data from a MEArec simulated data.

See:
https://mearec.readthedocs.io/en/latest/
https://github.com/alejoe91/MEArec
https://link.springer.com/article/10.1007/s12021-020-09467-7

Author : Alessio Buccino
"""

from copy import deepcopy

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class MEArecRawIO(BaseRawIO):
    """
    Class for "reading" simulated data from a MEArec file.

    Parameters
    ----------
    filename : str, default: ''
        The filename of the MEArec file to read.
    load_spiketrains : bool, default: True
        Whether or not to load spike train data.
    load_analogsignal : bool, default: True
        Whether or not to load continuous recording data.


    Examples
    --------
    >>> import neo.rawio
    >>> r = neo.rawio.MEArecRawIO(filename='mearec.h5', load_spiketrains=True)
    >>> r.parse_header()
    >>> print(r)
    >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                             seg_index=0,
                                             i_start=0,
                                             i_stop=1024,
                                             channel_names=channel_names)
    >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk,
                                                        dtype='float64',
                                                        channel_indexes=[0, 3, 6])
    >>> spike_timestamp = reader.spike_timestamps(unit_index=0, t_start=None, t_stop=None)
    >>> spike_times = reader.rescale_spike_timestamp(spike_timestamp, 'float64')

    """

    extensions = ["h5"]
    rawmode = "one-file"

    def __init__(self, filename="", load_spiketrains=True, load_analogsignal=True):
        BaseRawIO.__init__(self)
        self.filename = filename
        self.load_spiketrains = load_spiketrains
        self.load_analogsignal = load_analogsignal

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        load = ["channel_positions"]
        if self.load_analogsignal:
            load.append("recordings")
        if self.load_spiketrains:
            load.append("spiketrains")

        import MEArec as mr

        self._recgen = mr.load_recordings(
            recordings=self.filename, return_h5_objects=True, check_suffix=False, load=load, load_waveforms=False
        )

        self.info_dict = deepcopy(self._recgen.info)
        self.channel_positions = self._recgen.channel_positions
        if self.load_analogsignal:
            self._recordings = self._recgen.recordings
        if self.load_spiketrains:
            self._spiketrains = self._recgen.spiketrains

        self._sampling_rate = self.info_dict["recordings"]["fs"]
        self.duration_seconds = self.info_dict["recordings"]["duration"]
        self._num_frames = int(self._sampling_rate * self.duration_seconds)
        self._num_channels = self.channel_positions.shape[0]
        self._dtype = self.info_dict["recordings"]["dtype"]

        signals = [("Signals", "0")] if self.load_analogsignal else []
        signal_streams = np.array(signals, dtype=_signal_stream_dtype)

        sig_channels = []
        if self.load_analogsignal:
            for c in range(self._num_channels):
                ch_name = f"ch{c}"
                chan_id = str(c + 1)
                sr = self._sampling_rate  # Hz
                dtype = self._dtype
                units = "uV"
                gain = 1.0
                offset = 0.0
                stream_id = "0"
                sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))

        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # creating units channels
        spike_channels = []
        if self.load_spiketrains:
            for c in range(len(self._spiketrains)):
                unit_name = f"unit{c}"
                unit_id = f"#{c}"
                # if spiketrains[c].waveforms is not None:
                wf_units = ""
                wf_gain = 1.0
                wf_offset = 0.0
                wf_left_sweep = 0
                wf_sampling_rate = self._sampling_rate
                spike_channels.append(
                    (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                )

        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()
        for block_index in range(1):
            bl_ann = self.raw_annotations["blocks"][block_index]
            bl_ann["mearec_info"] = self.info_dict

    def _segment_t_start(self, block_index, seg_index):
        all_starts = [[0.0]]
        return all_starts[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._num_frames / self._sampling_rate
        all_stops = [[t_stop]]
        return all_stops[block_index][seg_index]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._num_frames

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        if not self.load_analogsignal:
            raise AttributeError("Recordings not loaded. Set load_analogsignal=True in MEArecRawIO constructor")

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._num_frames

        if channel_indexes is None:
            channel_indexes = slice(self._num_channels)
        if isinstance(channel_indexes, slice):
            raw_signals = self._recordings[i_start:i_stop, channel_indexes]
        else:
            # sort channels because h5py neeeds sorted indexes
            if np.any(np.diff(channel_indexes) < 0):
                sorted_channel_indexes = np.sort(channel_indexes)
                sorted_idx = np.array([list(sorted_channel_indexes).index(ch) for ch in channel_indexes])
                raw_signals = self._recordings[i_start:i_stop, sorted_channel_indexes]
                raw_signals = raw_signals[:, sorted_idx]
            else:
                raw_signals = self._recordings[i_start:i_stop, channel_indexes]
        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):

        return len(self._spiketrains[unit_index])

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):

        spike_timestamps = self._spiketrains[unit_index].times.magnitude
        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)
        timestamp_idxs = np.where((spike_timestamps >= t_start) & (spike_timestamps < t_stop))

        return spike_timestamps[timestamp_idxs]

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        return spike_timestamps.astype(dtype)

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None



"""
Demonstration of lazy load and proxy objects
============================================

"""

################################################
# Import our packages first
# It is often nice to have units so we will also
# import quantities

import urllib
import neo
import quantities as pq
import numpy as np


###############################################
# Let's get a file
# NeuralEnsemble maintains a wide variety of small test
# datasets that are free to use. We can use urllib to pull
# down one of these files for use

url_repo = "https://web.gin.g-node.org/NeuralEnsemble/ephy_testing_data/raw/master/"
# Get med file
distantfile = url_repo + "micromed/File_micromed_1.TRC"
localfile = "./File_micromed_1.TRC"
urllib.request.urlretrieve(distantfile, localfile)

##################################################
# create a reader
# creating a reader for neo is easy it just requires using
# the name of the desired reader and providing either a filename
# or a directory name (reader dependent). Since we got a micromed
# file we will use MicromedIO.

reader = neo.MicromedIO(filename="File_micromed_1.TRC")
reader.parse_header()

############################################################
# as always we can look view some interesting information about the
# metadata and structure of a file just by printing the reader and
# it's header
print(reader)
print(f"Header information: {reader.header}")


#####################################################
# Now let's make a function that we want to apply to
# look at lazy vs eager uses of the API


def apply_my_fancy_average(sig_list):
    """basic average along triggers and then channels
    here we go back to numpy with magnitude
    to be able to use np.stack.

    Because neo uses quantities to keep track of units
    we can always get just the magnitude of an array
    with `.magnitude`
    """
    sig_list = [s.magnitude for s in sig_list]
    sigs = np.stack(sig_list, axis=0)
    return np.mean(np.mean(sigs, axis=0), axis=1)


#################################################
# Let's set our limits for both cases. We will
# use quantities to include time dimensions.

lim_start = -20 * pq.ms  # 20 milliseconds before
lim_end = +20 * pq.ms  # 20 milliseconds after

##################################################
# We start with eager (where `lazy=False`.) Everything
# is loaded into memory. We will read a segment of data.
# This includes analog signal data and events data
# (final contents of a segment are dependent on the
# underlying IO being used)


seg = reader.read_segment(lazy=False)
triggers = seg.events[0]
anasig = seg.analogsignals[0]  # here anasig contain the whole recording in memory
all_sig_chunks = []
for t in triggers.times:
    t0, t1 = (t + lim_start), (t + lim_end)
    anasig_chunk = anasig.time_slice(t0, t1)
    all_sig_chunks.append(anasig_chunk)

# After pulling all data into memory and then iterating through triggers
# we end by doing our average
m1 = apply_my_fancy_average(all_sig_chunks)

#####################################################
# Here we do `lazy=True`, i.e. we do lazy loading. We
# only load the data that we want into memory
# and we use a proxy object for our analogsignal until we
# load it chunk by chunk (no running out of memory!)

seg = reader.read_segment(lazy=True)
triggers = seg.events[0].load(time_slice=None)  # this load all triggers in memory
anasigproxy = seg.analogsignals[0]  # this is a proxy
all_sig_chunks = []
for t in triggers.times:
    t0, t1 = (t + lim_start), (t + lim_end)
    # at this step we load actual data into memory, but notice that we only load one
    # chunk of data at a time, so we reduce the memory strain
    anasig_chunk = anasigproxy.load(time_slice=(t0, t1))  # here real data are loaded
    all_sig_chunks.append(anasig_chunk)

# Finally we apply the same average as we did above
m2 = apply_my_fancy_average(all_sig_chunks)

##########################################################
# We see that either way the result is the same, but
# we do not exhaust our RAM/memory
print(f"Eagerly loading data and averaging: {m1}")
print(f"Lazy loading data and average {m2}")



"""
Tests of the neo.core.event.Event class
"""

import unittest
import warnings
from copy import deepcopy

import numpy as np
import quantities as pq
import pickle
import os
from numpy.testing import assert_array_equal

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.dataobject import ArrayDict
from neo.core.event import Event
from neo.core.epoch import Epoch
from neo.core import Segment
from neo.test.tools import (
    assert_neo_object_is_compliant,
    assert_arrays_equal,
    assert_arrays_almost_equal,
    assert_same_sub_schema,
    assert_same_attributes,
    assert_same_annotations,
    assert_same_array_annotations,
)

warnings.simplefilter("always")


class TestEvent(unittest.TestCase):

    def setUp(self):
        self.params = {"test2": "y1", "test3": True}
        self.arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        self.seg = Segment()
        self.evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=self.arr_ann,
            **self.params,
        )
        self.evt.annotate(test1=1.1, test0=[1, 2])
        self.evt.segment = self.seg

    def test_setup_compliant(self):
        assert_neo_object_is_compliant(self.evt)

    def test_Event_creation(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"names": ["a", "b", "c"], "index": np.arange(10, 13)}
        evt = Event(
            [1.1, 1.5, 1.7] * pq.ms,
            labels=np.array(["test event 1", "test event 2", "test event 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        assert_arrays_equal(evt.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_arrays_equal(evt.labels, np.array(["test event 1", "test event 2", "test event 3"], dtype="U"))
        self.assertEqual(evt.name, "test")
        self.assertEqual(evt.description, "tester")
        self.assertEqual(evt.file_origin, "test.file")
        self.assertEqual(evt.annotations["test0"], [1, 2])
        self.assertEqual(evt.annotations["test1"], 1.1)
        self.assertEqual(evt.annotations["test2"], "y1")
        self.assertTrue(evt.annotations["test3"])
        assert_arrays_equal(evt.array_annotations["names"], np.array(["a", "b", "c"]))
        assert_arrays_equal(evt.array_annotations["index"], np.arange(10, 13))
        self.assertIsInstance(evt.array_annotations, ArrayDict)

    def test_Event_invalid_times_dimension(self):
        data2d = np.array([1, 2, 3, 4]).reshape((4, -1))
        self.assertRaises(ValueError, Event, times=data2d * pq.s)

    def test_Event_creation_invalid_labels(self):
        self.assertRaises(ValueError, Event, [1.1, 1.5, 1.7] * pq.ms, labels=["A", "B"])

    def test_Event_creation_from_lists(self):
        evt = Event([1.1, 1.5, 1.7], ["test event 1", "test event 2", "test event 3"], units=pq.ms)
        assert_arrays_equal(evt.times, [1.1, 1.5, 1.7] * pq.ms)
        assert_arrays_equal(evt.labels, np.array(["test event 1", "test event 2", "test event 3"]))

    def tests_time_slice(self):

        evt = self.evt

        targ = Event([2.2, 2.9, 3.0] * pq.ms)
        result = evt.time_slice(t_start=2.0, t_stop=3.0)

        assert_arrays_equal(targ, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(5, 8))
        assert_arrays_equal(result.array_annotations["test"], np.arange(105, 108))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def tests_time_slice_deepcopy_annotations(self):
        params = {"test0": "y1", "test1": ["deeptest"], "test2": True}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            **params,
        )
        result = evt.time_slice(t_start=2.0, t_stop=3.0)
        evt.annotate(test0="y2", test2=False)
        evt.annotations["test1"][0] = "shallowtest"

        self.assertNotEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(evt.annotations["test2"], result.annotations["test2"])

    def test__time_slice_deepcopy_annotations(self):
        params1 = {"test0": "y1", "test1": ["deeptest"], "test2": True}
        self.evt.annotate(**params1)
        # time_slice spike train, keep sliced spike times
        t_start = 2.1 * pq.ms
        t_stop = 3.05 * pq.ms
        result = self.evt.time_slice(t_start, t_stop)

        # Change annotations of original
        params2 = {"test0": "y2", "test2": False}
        self.evt.annotate(**params2)
        self.evt.annotations["test1"][0] = "shallowtest"

        self.assertNotEqual(self.evt.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.evt.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.evt.annotations["test2"], result.annotations["test2"])

        # Change annotations of result
        params3 = {"test0": "y3"}
        result.annotate(**params3)
        result.annotations["test1"][0] = "shallowtest2"

        self.assertNotEqual(self.evt.annotations["test0"], result.annotations["test0"])
        self.assertNotEqual(self.evt.annotations["test1"], result.annotations["test1"])
        self.assertNotEqual(self.evt.annotations["test2"], result.annotations["test2"])

    def test__time_slice_deepcopy_array_annotations(self):
        length = self.evt.shape[-1]
        params1 = {
            "test0": [f"y{i}" for i in range(length)],
            "test1": ["deeptest" for i in range(length)],
            "test2": [(-1) ** i > 0 for i in range(length)],
        }
        self.evt.array_annotate(**params1)
        # time_slice spike train, keep sliced spike times
        t_start = 2.1 * pq.ms
        t_stop = 3.05 * pq.ms
        result = self.evt.time_slice(t_start, t_stop)

        # Change annotations of original
        params2 = {
            "test0": [f"x{i}" for i in range(length)],
            "test2": [(-1) ** (i + 1) > 0 for i in range(length)],
        }
        self.evt.array_annotate(**params2)
        self.evt.array_annotations["test1"][6] = "shallowtest"

        self.assertFalse(all(self.evt.array_annotations["test0"][5:8] == result.array_annotations["test0"]))
        self.assertFalse(all(self.evt.array_annotations["test1"][5:8] == result.array_annotations["test1"]))
        self.assertFalse(all(self.evt.array_annotations["test2"][5:8] == result.array_annotations["test2"]))

        # Change annotations of result
        params3 = {"test0": [f"z{i}" for i in range(5, 8)]}
        result.array_annotate(**params3)
        result.array_annotations["test1"][1] = "shallow2"

        self.assertFalse(all(self.evt.array_annotations["test0"][5:8] == result.array_annotations["test0"]))
        self.assertFalse(all(self.evt.array_annotations["test1"][5:8] == result.array_annotations["test1"]))
        self.assertFalse(all(self.evt.array_annotations["test2"][5:8] == result.array_annotations["test2"]))

    def test__time_slice_deepcopy_data(self):
        result = self.evt.time_slice(None, None)

        # Change values of original array
        self.evt[2] = 7.3 * self.evt.units

        self.assertFalse(all(self.evt == result))

        # Change values of sliced array
        result[3] = 9.5 * result.units

        self.assertFalse(all(self.evt == result))

    def test_time_slice_out_of_boundries(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        targ = evt
        result = evt.time_slice(t_start=0.0001, t_stop=30.0)

        assert_arrays_equal(targ, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(10))
        assert_arrays_equal(result.array_annotations["test"], np.arange(100, 110))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_empty(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.array([]), "test": np.array([])}
        evt = Event(
            [] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        result = evt.time_slice(t_start=0.0001, t_stop=30.0)
        assert_neo_object_is_compliant(evt)

        assert_arrays_equal(evt, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.asarray([]))
        assert_arrays_equal(result.array_annotations["test"], np.asarray([]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_stop(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        targ = Event([2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms)
        assert_neo_object_is_compliant(evt)

        t_start = 2.0
        t_stop = None
        result = evt.time_slice(t_start, t_stop)

        assert_arrays_equal(targ, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(5, 10))
        assert_arrays_equal(result.array_annotations["test"], np.arange(105, 110))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_start(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        targ = Event([0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0] * pq.ms)
        t_start = None
        t_stop = 3.0
        result = evt.time_slice(t_start, t_stop)

        assert_arrays_equal(targ, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(8))
        assert_arrays_equal(result.array_annotations["test"], np.arange(100, 108))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_none_both(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        assert_neo_object_is_compliant(evt)

        evt.annotate(test1=1.1, test0=[1, 2])
        t_start = None
        t_stop = None
        result = evt.time_slice(t_start, t_stop)

        assert_arrays_equal(evt, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(10))
        assert_arrays_equal(result.array_annotations["test"], np.arange(100, 110))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_time_slice_differnt_units(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(9), "test": np.arange(100, 109)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        assert_neo_object_is_compliant(evt)
        evt.annotate(test1=1.1, test0=[1, 2])

        targ = Event([2.2, 2.9] * pq.ms, name="test", description="tester", file_origin="test.file", test1=1, **params)
        assert_neo_object_is_compliant(targ)
        targ.annotate(test1=1.1, test0=[1, 2])

        t_start = 0.002 * pq.s
        t_stop = 0.003 * pq.s

        result = evt.time_slice(t_start, t_stop)

        assert_arrays_equal(targ, result)
        self.assertEqual(targ.name, result.name)
        self.assertEqual(targ.description, result.description)
        self.assertEqual(targ.file_origin, result.file_origin)
        self.assertEqual(targ.annotations["test0"], result.annotations["test0"])
        self.assertEqual(targ.annotations["test1"], result.annotations["test1"])
        self.assertEqual(targ.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(5, 7))
        assert_arrays_equal(result.array_annotations["test"], np.arange(105, 107))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test__time_slice_should_set_parents_to_None(self):
        # When timeslicing, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.evt.time_slice(1 * pq.ms, 3 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__deepcopy_should_set_parents_objects_to_None(self):
        # Deepcopy should destroy references to parents
        result = deepcopy(self.evt)
        self.assertEqual(result.segment, None)

    def test_slice(self):
        params = {"test2": "y1", "test3": True}
        arr_ann = {"index": np.arange(10), "test": np.arange(100, 110)}
        evt = Event(
            [0.1, 0.5, 1.1, 1.5, 1.7, 2.2, 2.9, 3.0, 3.1, 3.3] * pq.ms,
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            array_annotations=arr_ann,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        targ = Event([2.2, 2.9, 3.0] * pq.ms)
        result = evt[5:8]

        assert_arrays_equal(targ, result)
        self.assertEqual(evt.name, result.name)
        self.assertEqual(evt.description, result.description)
        self.assertEqual(evt.file_origin, result.file_origin)
        self.assertEqual(evt.annotations["test0"], result.annotations["test0"])
        self.assertEqual(evt.annotations["test1"], result.annotations["test1"])
        self.assertEqual(evt.annotations["test2"], result.annotations["test2"])
        assert_arrays_equal(result.array_annotations["index"], np.arange(5, 8))
        assert_arrays_equal(result.array_annotations["test"], np.arange(105, 108))
        self.assertIsInstance(result.array_annotations, ArrayDict)

    def test_Event_repr(self):
        params = {"test2": "y1", "test3": True}
        evt = Event(
            [1.1, 1.5, 1.7] * pq.ms,
            labels=np.array(["test event 1", "test event 2", "test event 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        targ = "<Event: test event 1@1.1 ms, test event 2@1.5 ms, " + "test event 3@1.7 ms>"

        res = repr(evt)

        self.assertEqual(targ, res)

    def test_Event_merge(self):
        warnings.simplefilter("always")
        params1 = {"test2": "y1", "test3": True}
        params2 = {"test2": "no", "test4": False}
        paramstarg = {"test2": "yes;no", "test3": True, "test4": False}
        arr_ann1 = {"index": np.arange(10, 13)}
        arr_ann2 = {"index": np.arange(3), "test": ["a", "b", "c"]}
        evt1 = Event(
            [1.1, 1.5, 1.7] * pq.ms,
            labels=np.array(["test event 1 1", "test event 1 2", "test event 1 3"], dtype="U"),
            name="test",
            description="tester 1",
            file_origin="test.file",
            array_annotations=arr_ann1,
            test1=1,
            **params1,
        )
        evt2 = Event(
            [2.1, 2.5, 2.7] * pq.us,
            labels=np.array(["test event 2 1", "test event 2 2", "test event 2 3"], dtype="U"),
            name="test",
            description="tester 2",
            file_origin="test.file",
            array_annotations=arr_ann2,
            test1=1,
            **params2,
        )
        evttarg = Event(
            [1.1, 1.5, 1.7, 0.0021, 0.0025, 0.0027] * pq.ms,
            labels=np.array(
                [
                    "test event 1 1",
                    "test event 1 2",
                    "test event 1 3",
                    "test event 2 1",
                    "test event 2 2",
                    "test event 2 3",
                ],
                dtype="U",
            ),
            name="test",
            description="merge(tester 1, tester 2)",
            file_origin="test.file",
            array_annotations={"index": [10, 11, 12, 0, 1, 2]},
            test1=1,
            **paramstarg,
        )
        assert_neo_object_is_compliant(evt1)
        assert_neo_object_is_compliant(evt2)
        assert_neo_object_is_compliant(evttarg)

        with warnings.catch_warnings(record=True) as w:
            evtres = evt1.merge(evt2)

            self.assertEqual(len(w), 1)
            self.assertEqual(w[0].category, UserWarning)
            self.assertSequenceEqual(
                str(w[0].message),
                "The following array annotations were "
                "omitted, because they were only present"
                " in one of the merged objects: "
                "[] from the one that was merged "
                "into and ['test'] from the one that "
                "was merged into the other",
            )

        assert_neo_object_is_compliant(evtres)
        assert_same_sub_schema(evttarg, evtres)
        # Remove this, when array_annotations are added to assert_same_sub_schema
        assert_arrays_equal(evtres.array_annotations["index"], np.array([10, 11, 12, 0, 1, 2]))
        self.assertTrue("test" not in evtres.array_annotations)
        self.assertIsInstance(evtres.array_annotations, ArrayDict)

    def test_set_labels(self):
        evt = Event([1.1, 1.5, 1.7] * pq.ms, labels=["A", "B", "C"])
        assert_array_equal(evt.labels, np.array(["A", "B", "C"]))
        evt.labels = ["D", "E", "F"]
        assert_array_equal(evt.labels, np.array(["D", "E", "F"]))
        self.assertRaises(ValueError, setattr, evt, "labels", ["X", "Y"])

    def test__children(self):
        params = {"test2": "y1", "test3": True}
        evt = Event(
            [1.1, 1.5, 1.7] * pq.ms,
            labels=np.array(["test event 1", "test event 2", "test event 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
            test1=1,
            **params,
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        segment = Segment(name="seg1")
        segment.events = [evt]
        segment.check_relationships()

        self.assertEqual(evt._parent_objects, ("Segment",))

        self.assertEqual(evt._parent_containers, ("segment",))

        self.assertEqual(evt._parent_objects, ("Segment",))
        self.assertEqual(evt._parent_containers, ("segment",))

        self.assertEqual(len(evt.parents), 1)
        self.assertEqual(evt.parents[0].name, "seg1")

        assert_neo_object_is_compliant(evt)

    @unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
    def test__pretty(self):
        evt = Event(
            [1.1, 1.5, 1.7] * pq.ms,
            labels=np.array(["test event 1", "test event 2", "test event 3"], dtype="U"),
            name="test",
            description="tester",
            file_origin="test.file",
        )
        evt.annotate(test1=1.1, test0=[1, 2])
        assert_neo_object_is_compliant(evt)

        prepr = pretty(evt)
        targ = (
            f"Event containing {len(evt)} events with labels; "
            f"time units {evt.units.dimensionality.string}; datatype float64 "
            f"\nname: '{evt.name}'\ndescription: '{evt.description}'"
            f"\nannotations: {pretty(evt.annotations)}"
        )

        self.assertEqual(prepr, targ)

    def test__time_slice(self):
        data = [2, 3, 4, 5] * pq.ms
        evt = Event(data, foo="bar")

        evt1 = evt.time_slice(2.2 * pq.ms, 4.2 * pq.ms)
        assert_arrays_equal(evt1.times, [3, 4] * pq.ms)
        self.assertEqual(evt.annotations, evt1.annotations)

        evt2 = evt.time_slice(None, 4.2 * pq.ms)
        assert_arrays_equal(evt2.times, [2, 3, 4] * pq.ms)

        evt3 = evt.time_slice(2.2 * pq.ms, None)
        assert_arrays_equal(evt3.times, [3, 4, 5] * pq.ms)

    def test__time_shift_same_attributes(self):
        result = self.evt.time_shift(1 * pq.ms)
        assert_same_attributes(result, self.evt, exclude=["times"])

    def test__time_shift_same_annotations(self):
        result = self.evt.time_shift(1 * pq.ms)
        assert_same_annotations(result, self.evt)

    def test__time_shift_same_array_annotations(self):
        result = self.evt.time_shift(1 * pq.ms)
        assert_same_array_annotations(result, self.evt)

    def test__time_shift_should_set_parents_to_None(self):
        # When time-shifting, a deep copy is made,
        # thus the reference to parent objects should be destroyed
        result = self.evt.time_shift(1 * pq.ms)
        self.assertEqual(result.segment, None)

    def test__time_shift_by_zero(self):
        shifted = self.evt.time_shift(0 * pq.ms)
        assert_arrays_equal(shifted.times, self.evt.times)

    def test__time_shift_same_units(self):
        shifted = self.evt.time_shift(10 * pq.ms)
        assert_arrays_equal(shifted.times, self.evt.times + 10 * pq.ms)

    def test__time_shift_different_units(self):
        shifted = self.evt.time_shift(1 * pq.s)
        assert_arrays_equal(shifted.times, self.evt.times + 1000 * pq.ms)

    def test_as_array(self):
        data = [2, 3, 4, 5]
        evt = Event(data * pq.ms)
        evt_as_arr = evt.as_array()
        self.assertIsInstance(evt_as_arr, np.ndarray)
        assert_array_equal(data, evt_as_arr)

    def test_as_quantity(self):
        data = [2, 3, 4, 5]
        evt = Event(data * pq.ms)
        evt_as_q = evt.as_quantity()
        self.assertIsInstance(evt_as_q, pq.Quantity)
        assert_array_equal(data * pq.ms, evt_as_q)

    def test_to_epoch(self):
        seg = Segment(name="test")
        event = Event(times=np.array([5.0, 12.0, 23.0, 45.0]), units="ms", labels=np.array(["A", "B", "C", "D"]))
        event.segment = seg

        # Mode 1
        epoch = event.to_epoch()
        self.assertIsInstance(epoch, Epoch)
        assert_array_equal(epoch.times.magnitude, np.array([5.0, 12.0, 23.0]))
        assert_array_equal(epoch.durations.magnitude, np.array([7.0, 11.0, 22.0]))
        assert_array_equal(epoch.labels, np.array(["A-B", "B-C", "C-D"]))

        # Mode 2
        epoch = event.to_epoch(pairwise=True)
        assert_array_equal(epoch.times.magnitude, np.array([5.0, 23.0]))
        assert_array_equal(epoch.durations.magnitude, np.array([7.0, 22.0]))
        assert_array_equal(epoch.labels, np.array(["A-B", "C-D"]))

        # Mode 3 (scalar)
        epoch = event.to_epoch(durations=2.0 * pq.ms)
        assert_array_equal(epoch.times.magnitude, np.array([5.0, 12.0, 23.0, 45.0]))
        assert_array_equal(epoch.durations.magnitude, np.array([2.0, 2.0, 2.0, 2.0]))
        self.assertEqual(epoch.durations.size, 4)
        assert_array_equal(epoch.labels, np.array(["A", "B", "C", "D"]))

        # Mode 3 (array)
        epoch = event.to_epoch(durations=np.array([2.0, 3.0, 4.0, 5.0]) * pq.ms)
        assert_array_equal(epoch.times.magnitude, np.array([5.0, 12.0, 23.0, 45.0]))
        assert_array_equal(epoch.durations.magnitude, np.array([2.0, 3.0, 4.0, 5.0]))
        assert_array_equal(epoch.labels, np.array(["A", "B", "C", "D"]))

        # Error conditions
        self.assertRaises(ValueError, event.to_epoch, pairwise=True, durations=2.0 * pq.ms)

        odd_event = Event(times=np.array([5.0, 12.0, 23.0]), units="ms", labels=np.array(["A", "B", "C"]))
        self.assertRaises(ValueError, odd_event.to_epoch, pairwise=True)

        # todo: fix Epoch, as the following does not raise a ValueError  # self.assertRaises(
        # ValueError, event.to_epoch, durations=2.0)  # missing units

    def test_rescale(self):
        times = [2, 3, 4, 5]
        labels = ["A", "B", "C", "D"]
        arr_ann = {"index": np.arange(4), "test": ["a", "b", "c", "d"]}
        evt = Event(times * pq.ms, labels=labels, array_annotations=arr_ann)
        result = evt.rescale(pq.us)

        self.assertIsInstance(result, Event)
        assert_neo_object_is_compliant(result)
        assert_arrays_equal(result.array_annotations["index"], np.arange(4))
        assert_arrays_equal(result.array_annotations["test"], np.array(["a", "b", "c", "d"]))
        self.assertIsInstance(result.array_annotations, ArrayDict)

        self.assertEqual(result.units, 1 * pq.us)
        assert_array_equal(evt.labels, result.labels)
        assert_arrays_almost_equal(result.times, [2000, 3000, 4000, 5000] * pq.us, 1e-9)
        assert_arrays_almost_equal(result.times.magnitude, np.array([2000, 3000, 4000, 5000]), 1e-9)


class TestDuplicateWithNewData(unittest.TestCase):
    def setUp(self):
        self.data = np.array([0.1, 0.5, 1.2, 3.3, 6.4, 7])
        self.dataquant = self.data * pq.ms
        self.arr_ann = {"index": np.arange(6), "test": ["a", "b", "c", "d", "e", "f"]}
        self.event = Event(
            times=self.dataquant, labels=np.array(["a", "b", "c", "d", "e", "f"]), array_annotations=self.arr_ann
        )

    def test_duplicate_with_new_data(self):
        signal1 = self.event
        new_times = np.sort(np.random.uniform(0, 100, (self.event.size))) * pq.ms
        new_labels = np.array(list("zyxwvutsrqponmlkjihgfedcba"[: self.event.size]))
        signal1b = signal1.duplicate_with_new_data(new_times, new_labels)
        assert_arrays_almost_equal(np.asarray(signal1b), np.asarray(new_times), 1e-12)
        assert_arrays_equal(signal1b.labels, new_labels)
        # After duplicating, array annotations should always be empty,
        # because different length of data would cause inconsistencies
        self.assertTrue("index" not in signal1b.array_annotations)
        self.assertTrue("test" not in signal1b.array_annotations)
        self.assertIsInstance(signal1b.array_annotations, ArrayDict)


class TestEventFunctions(unittest.TestCase):
    def test__pickle(self):
        arr_ann = {"index": np.arange(3), "test": ["a", "b", "c"]}
        event1 = Event(
            np.arange(0, 30, 10) * pq.s,
            labels=np.array(["t0", "t1", "t2"], dtype="U"),
            units="s",
            annotation1="foo",
            annotation2="bar",
            array_annotations=arr_ann,
        )
        fobj = open("./pickle", "wb")
        pickle.dump(event1, fobj)
        fobj.close()

        fobj = open("./pickle", "rb")
        try:
            event2 = pickle.load(fobj)
        except ValueError:
            event2 = None

        fobj.close()
        assert_array_equal(event1.times, event2.times)
        assert_arrays_equal(event2.array_annotations["index"], np.array(arr_ann["index"]))
        assert_arrays_equal(event2.array_annotations["test"], np.array(arr_ann["test"]))
        self.assertIsInstance(event2.array_annotations, ArrayDict)
        # Make sure the dict can perform correct checks after unpickling
        event2.array_annotations["anno3"] = list(range(3, 6))
        with self.assertRaises(ValueError):
            event2.array_annotations["anno4"] = [2, 1]
        os.remove("./pickle")
        self.assertEqual(event2.annotations, event1.annotations)


if __name__ == "__main__":
    unittest.main()
